# Import external dependencies
import numpy as np
from skimage import restoration as rest
from skimage import segmentation as seg
from skimage import filters as filt
from skimage import morphology as morph
import skimage.feature as sfeature
from skimage.util import img_as_bool, img_as_ubyte, img_as_float

# Import local packages
from . import ski_interactive_processing as ifun


def interact_driver_skeletonize(img_in, fltr_name_in="scikit"):
    """
    Binarizes an image by thinning connected pixels until they are just
    one pixel wide, also known as skeletonization. The implementation
    is done by SciKit-Image: skimage.morphology.skeletonize(). This is
    an interactive function that enables the user to change the
    parameters of the filter and see the results, thanks to
    the "widgets" available in Matplotlib.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: Numpy array for a binary image. It is assumed that the
        image is already segmented and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    fltr_name_in: A string that represents the name of the skeletonize
        algorithm to be applied. Currently, only one exists, named
        "scikit":

            "scikit": Uses functions from SciKit-Image to skeletonize
            the binary image.

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same data type as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.

    """
    
    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()

    # List of strings of supported filter types
    fltr_list = ["scikit"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported algorithm for skeletonizing an image."\
            f"Supported algorithms are: \n  {fltr_list}"\
            f"\nDefaulting to Scikit-Image's skeletonization.")
        fltr_name = "scikit"

    if fltr_name == "scikit":
        [img_fltr, fltr_params] = ifun.interact_skeletonize(img_in)
    
    # Using this function just to write to standard output
    apply_driver_skeletonize(img_in, fltr_params)

    return img_fltr


def apply_driver_skeletonize(img_in, fltr_params_in, quiet_in=False):
    """
    Binarizes an image by thinning connected pixels until they are just
    one pixel wide, also known as skeletonization. The implementation
    is done by SciKit-Image: skimage.morphology.skeletonize(). This is
    an non-interactive function related to the above function,
    interact_driver_skeletonize().

    ---- INPUT ARGUMENTS ---- 
    [img_in]: Numpy array for a binary image. It is assumed that the
        image is already segmented and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    fltr_params_in: A list of parameters needed to perform the 
        skeletonize operation. The first parameter is a string, which
        determines what type of skeletonize algorithm to be applied, as
        well as the definitions of the remaining parameters. Currently,
        the only option is "scikit":

            ["scikit", apply_skel]
                apply_skel: A boolean which applies the skeletonize 
                    algorithm if True, and does not modify the image
                    if False.

    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image after performing the
        image processing procedures. img_out is in the same data type as
        img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    if fltr_name == "scikit":
        apply_skel = fltr_params[1]

        if apply_skel:
            img_bool = img_as_bool(img)
            img_temp = morph.skeletonize(img_bool)
            img_fltr = img_as_ubyte(img_temp)

        else:
            img_fltr = img

        if not quiet:
            print(f"\nSuccessfully skeletonized the image:\n"\
                f"    Applied skeletonize: {apply_skel}")

    return img_fltr


def interact_driver_del_features(img_in, fltr_name_in="scikit"):
    """
    Interactively delete features, and/or fill holes, of a specified
    size (i.e., area). This is accomplished using SkiKit-Image:
    skimage.morphology.remove_small_holes() and
    skimage.morphology.remove_small_objects(). This is an interactive
    function that enables the user to change the parameters of the
    filter and see the results, thanks to the "widgets" available in
    Matplotlib.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: Numpy array for a binary image. It is assumed that the
        image is already segmented and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.

    fltr_name_in: A string that represents the name of the deletion
        algorithm to be applied. Currently, only one exists, named
        "scikit": 

            "scikit": Uses functions from SciKit-Image to label and
            delete features of a specified size. Also uses functions
            from SciKit-Image to fill in holes of a specified size.
            1-connectivity is used in all cases.

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same data type as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()

    # List of strings of supported filter types
    fltr_list = ["scikit"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported algorithm for deleting features."\
            f"Supported algorithms are: \n  {fltr_list}"\
            f"\nDefaulting to Scikit-Image's feature deletion.")
        fltr_name = "scikit"

    if fltr_name == "scikit":
        [img_fltr, fltr_params] = ifun.interact_del_features(img_in)

    # Using this function just to write to standard output
    apply_driver_del_features(img_in, fltr_params)

    return img_fltr


def apply_driver_del_features(img_in, fltr_params_in, quiet_in=False):
    """
    Delete features and/or fill holes of a specified size (i.e., area).
    This is accomplished using the SkiKit-Image library. This is the
    non-interactive version of interact_driver_del_features() defined
    above.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: Numpy array for a binary image. It is assumed that the
        image is already segmented and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.   

    fltr_params_in: A list of parameters needed to perform the fill and
        deletion operations. The first parameter is a string, which
        determines what type of feature detection algorithms to use, as
        well as the definitions of the remaining parameters. Currently,
        only the "scikit" implementation is complete:

            ["scikit", max_hole_sz, min_feat_sz]
                max_hole_sz: The maximum area, in pixels, of a 
                    contiguous hole that will be filled. 1-connectivity
                    is assumed, and this should be an integer.
                min_feat_sz: The smallest allowable object size, in
                    pixels, assuming 1-connectivity. This should be an
                    integer.

    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image after performing the
        image processing procedures. img_out is in the same data type as
        img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    if fltr_name == "scikit":
        max_hole_sz = fltr_params[1]
        min_feat_sz = fltr_params[2]

        img_bool = img_as_bool(img)
        img_temp = morph.remove_small_holes(img_bool, max_hole_sz, connectivity=1)
        img_fltr = morph.remove_small_objects(img_temp, min_feat_sz, connectivity=1)
        img_fltr = img_as_ubyte(img_fltr)

        if not quiet:
            print(f"\nSuccessfully removed small holes and small objects:\n"\
                f"    Minimum Allowable Feature Size: {min_feat_sz}\n"\
                f"    Maximum Allowable Hole Size: {max_hole_sz}")

    return img_fltr


def interact_driver_edge_detect(img_in, fltr_name_in):
    """
    Enhances the edges of an image which are detected using the Canny
    edge filter, implemented by SciKit-Image. This is an interactive
    function that enables the user to change the parameters of the
    filter and see the results, thanks to the "widgets" available in
    Matplotlib.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_name_in: A string that represents the name of the thresholding
        filter to be applied. Can be either: 

            "canny": Applies the Canny edge detection algorithm.

    ---- RETURNED ----
    [img_fltr]: Returns the final image after closing the interactive
        session. img_fltr is in the same format as img_in.

    [img_mask]: A binary image where white pixels represent the edges
        found by the Canny algorithm.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """
    
    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()

    # List of strings of supported filter types
    fltr_list = ["canny"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported edge detection. Supported "\
            f"edge detection algorithms are: \n  {fltr_list}"\
            "\nDefaulting to canny edge detection.")
        fltr_name = "canny"

    if fltr_name == "canny":
        [img_fltr, img_mask, fltr_params] = ifun.interact_canny_edge(img_in)

    # Using this function just to write to standard output
    apply_driver_edge_detect(img_in, fltr_params)

    return img_fltr, img_mask


def apply_driver_edge_detect(img_in, fltr_params_in, quiet_in=False):
    """
    Enhances the edges of an image which are detected using the Canny
    edge filter, implemented by SciKit-Image. This is the
    non-interactive version of interact_driver_edge_detect().

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.

    fltr_params_in: A list of parameters needed to perform the edge
        detection. The first parameter is a string, which determines
        what type of edge detection filter to be applied, as well as the 
        definitions of the remaining parameters. Only the "canny" option
        is implemented at the moment:

            ["canny", sigma_out, low_thresh_out, high_thresh_out]
                sigma_out: Standard deviation of the Gaussian filter, 
                    which should be a float.
                low_thresh_out: Lower bound for hysteresis thresholding
                    (linking edges), which should be a float.
                high_thresh_out: Upper bound for hysteresis thresholding
                    (linking edges), which should be a float.

    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_fltr]: Returns the final image after closing the interactive
        session. img_fltr is in the same format as img_in.

    [img_mask]: A binary image where white pixels represent the edges
        found by the Canny algorithm.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    if fltr_name == "canny":
        sigma_out = fltr_params[1]
        low_thresh_out = fltr_params[2] 
        high_thresh_out = fltr_params[3]

        img_mask = sfeature.canny(img, sigma=sigma_out, 
            low_threshold=low_thresh_out, high_threshold=high_thresh_out)

        mask_out = img_as_ubyte(img_mask)
        img_fltr = img.copy()
        img_fltr[img_mask] = img[img_mask]/2

        if not quiet:
            print(f"\nSuccessfully applied the 'canny' edge detection:\n"\
                f"    Sigma: {sigma_out}\n"\
                f"    Low Threshold: {low_thresh_out}\n"\
                f"    High Threshold: {high_thresh_out}")

    return img_fltr, mask_out


def interact_driver_thresholding(img_in, fltr_name_in):
    """
    Interactively implements a thresholding operation in order to
    segment (or binarize) a UINT8 image into black (0) and white
    (255) pixels. The output parameters of this function are suitable
    for input for apply_driver_thresholding() below.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_name_in: A string that represents the name of the thresholding
        filter to be applied. Can be either: 
            
            "hysteresis_threshold_slider": Applies a threshold to an 
            image using hysteresis thresholding, which considers the
            connectivity of features. This interactive GUI utilizes
            slider widgets for input.
            
            "hysteresis_threshold_text": Applies a threshold to an 
            image using hysteresis thresholding, which considers the
            connectivity of features. This interactive GUI utilizes
            text box widgets for input.

            "adaptive_threshold": Applies adaptive thresholding, also
            known as local thresholding.

            "global_threshold": Uses a single value of grayscale 
            intensity as the thresholding criterion for the image.

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same format as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()

    # List of strings of supported filter types
    fltr_list = ["hysteresis_threshold_slider", "hysteresis_threshold_text",
        "global_threshold"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported threshold type. Supported "\
            f"threshold types are: \n  {fltr_list}"\
            "\nDefaulting to hysteresis thresholding.")
        fltr_name = "hysteresis_threshold_slider"

    # Filter Type: "hysteresis_threshold" for slider inputs
    if fltr_name == "hysteresis_threshold_slider":
        [img_fltr, fltr_params] = ifun.interact_hysteresis_threshold(img_in)

    # Filter Type: "hysteresis_threshold2" for text inputs
    elif fltr_name == "hysteresis_threshold_text":
        [img_fltr, fltr_params] = ifun.interact_hysteresis_threshold2(img_in)

    elif fltr_name == "global_threshold":
        [img_fltr, fltr_params] = ifun.interact_global_thresholding(img_in)

    elif fltr_name == "adaptive_threshold":
        [img_fltr, fltr_params] = ifun.interact_adaptive_thresholding(img_in)

    # Using this function just to write to standard output
    apply_driver_thresholding(img_in, fltr_params)

    return img_fltr


def apply_driver_thresholding(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a thresholding operation in order to segment (or binarize) a
    UINT8 image into black (0) and white(255) pixels. This is the
    non-interactive version of interact_driver_thresholding().

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_params_in: A list of parameters needed to perform the threshold
        operation. The first parameter is a string, which determines
        what type of threshold filter to be applied, as well as the 
        definitions of the remaining parameters. Available options are: 
            
            ["hysteresis_threshold", low_val_out, high_val_out] 
                low_val_out: Lower threshold intensity as an integer
                high_val_out: Upper threshold intensity as an integer

            ["adaptive_threshold", block_sz, thresh_offset]
                block_sz: Odd size of pixel neighborhood which is used
                    to calculate the threshold value Should be an
                    integer.
                thresh_offset: Constant subtracted from weighted mean of
                    neighborhood to calculate the local threshold value.
                    Default offset is 0.0, and it is treated as a float.

            ["global_threshold", global_thresh]
                global_thresh: A grayscale intensity to use as a 
                    criterion for thresholding. Values greater than 
                    this threshold will become white. Values less than
                    this threshold will become black.
            
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image after performing the
        image processing procedures. img_out is in the same format as
        img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    if fltr_name == "hysteresis_threshold":
        low_val_in = fltr_params[1]
        high_val_in = fltr_params[2]

        img_temp = filt.apply_hysteresis_threshold(img, low_val_in, 
            high_val_in)
        img_fltr = img_as_ubyte(img_temp)
        
        if not quiet:
            print(f"\nSuccessfully applied the 'hysteresis_threshold':\n"\
                f"    Lower grayscale intensity limit: {low_val_in}\n"\
                f"    Upper grayscale intensity limit: {high_val_in}")

    elif fltr_name == "global_threshold":
        global_thresh = fltr_params[1]
        img_fltr = img_as_ubyte(img > global_thresh)

        if not quiet:
            print(f"\nSuccessfully applied the 'global_threshold':\n"\
                f"    Global threshold value: {global_thresh}")

    elif fltr_name == "adaptive_threshold":
        block_sz = fltr_params[1]
        thresh_offset = fltr_params[2]

        if block_sz % 2 == 0:
            block_sz = block_sz + 1

        img_thresh = filt.threshold_local(img, block_size=block_sz, 
            method='gaussian', offset=thresh_offset)

        img_fltr = img_as_ubyte(img > img_thresh)

        if not quiet:
            print(f"\nSuccessfully applied the 'adaptive_threshold':\n"\
                f"    Block Size: {block_sz}\n"\
                f"    Offset (+/-): {thresh_offset}")

    else:
        print(f"\nERROR: {fltr_name} is not a supported sharpening filter.")
        img_fltr = img

    return img_fltr


def interact_driver_sharpen(img_in, fltr_name_in):
    """
    Interactively implements a sharpening filter. The output parameters
    of this function are suitable for input for apply_driver_sharpen()
    below.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_name_in: A string that represents the name of the sharpening
        filter to be applied. Currently only "unsharp_mask" is 
        supported.  
            
            "unsharp_mask":  Sharpens an image based on an unsharp
                mask using a Ski-Image function. An unsharp mask is
                based on a weighted addition between  the original
                image and Gaussian blurred version. 

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same format as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()

    # List of strings of supported filter types
    fltr_list = ["unsharp_mask"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported filter type. Supported "\
            f"filter types are: \n  {fltr_list}"\
            "\nDefaulting to unsharp mask.")
        fltr_name = "unsharp_mask"

    # -------- Filter Type: "unsharp_mask" --------
    if fltr_name == "unsharp_mask":
        [img_fltr, fltr_params] = ifun.interact_unsharp_mask(img_in)

    # Using this function just to write to standard output
    apply_driver_sharpen(img_in, fltr_params)

    return img_fltr


def apply_driver_sharpen(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a sharpening filter. This is the non-interactive version of
    interact_driver_sharpen().

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_params_in: A list of parameters needed to perform the 
        sharpening operation. The first parameter is a string, which
        determines what type of sharpen filter to be applied, as well
        as the definitions of the remaining parameters. As of now,
        fltr_params_in[0] must be "unsharp_mask". Example parameter
        lists are given below for each type, 
            
            ["unsharp_mask", radius_out, amount_out] 

                radius_out: Radius of the kernel for the unsharp filter. 
                    If zero, then no filter was applied. Should be an
                    integer.

                amount_out: The sharpening details will be amplified 
                    with this factor, which can be a negative or 
                    positive float.
            
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.
            
    NOTE: Currently, only sharpening via an unsharp mask has been
    implemented for thresholding from the Skimage library. For
    additional methods, see the OpenCV interactive scripts in
    cv_driver_functions.py.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image after performing the
        image processing procedures. img_out is in the same format as
        img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    if fltr_name == "unsharp_mask":
        radius_in = fltr_params[1]
        amount_in = fltr_params[2]

        img_temp = filt.unsharp_mask(img, radius=radius_in, 
            amount=amount_in, channel_axis=None)
        img_fltr = img_as_ubyte(img_temp)
        
        if not quiet:
            print(f"\nSuccessfully applied the 'unsharp_mask' sharpening filter:\n"\
                f"    Radius of sharpening kernel: {radius_in}\n"\
                f"    Amount of sharpening: {amount_in}")

    else:
        print(f"\nERROR: {fltr_name} is not a supported sharpening filter.")
        img_fltr = img

    return img_fltr


def interact_driver_denoise(img_in, fltr_name_in):
    """
    Interactively implements a denoising filter. The output parameters
    of this function are suitable for input for apply_driver_denoise()
    below.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_name_in: A string that represents the name of the denoising
        filter to be applied. Currently, "tv_chambolle" or "nl_means" 
        are supported.  
            
            "tv_chambolle": Perform total-variation denoising on an 
                image. The principle of total variation denoising is to
                minimize the total variation of the image, which can be
                roughly described as the integral of the norm of the
                image gradient. Total variation denoising tends to
                produce “cartoon-like” images, that is, 
                piecewise-constant images.

            "nl_means": Perform non-local means denoising on an image. 
                The non-local means algorithm is well suited for
                denoising images with specific textures. The principle
                of the algorithm is to average the value of a given
                pixel with values of other pixels in a limited
                neighbourhood, provided that the patches centered on
                the other pixels are similar enough to the patch
                centered on the pixel of interest.

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same format as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()

    # List of strings of supported filter types
    fltr_list = ["tv_chambolle", "nl_means"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported filter type. Supported "\
            f"filter types are: \n  {fltr_list}"\
            "\nDefaulting to non-local means denoising.")
        fltr_name = "nl_means"

    # -------- Filter Type: "nl_means" --------
    if fltr_name == "nl_means":
        [img_fltr, fltr_params] = ifun.interact_nl_means_denoise(img_in)

    # -------- Filter Type: "tv_chambolle" --------
    elif fltr_name == "tv_chambolle":
        [img_fltr, fltr_params] = ifun.interact_tv_denoise(img_in)

    # Using this function just to write to standard output
    apply_driver_denoise(img_in, fltr_params)

    return img_fltr


def apply_driver_denoise(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a denoising filter. This is the non-interactive version of
    interact_driver_denoise().

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_params_in: A list of parameters needed to perform the 
        denoising operation. The first parameter is a string, which
        determines what type of denoise filter to be applied, as well
        as the definitions of the remaining parameters. As of now,
        fltr_params_in[0] must be either "nl_means" or "tv_chambolle". 
        Example parameter lists are given below for each type, 
            
            ["tv_chambolle", weight_out, eps_out, n_iter_max_out]

                weight_out: Denoising weight. The greater weight, the 
                    more denoising (at the expense of fidelity)

                eps_out: Relative difference of the value of the cost
                    function that determines the stop criterion. See the
                    Skimage documentation for additional details.

                n_iter_max_out: Maximal number of iterations used for 
                    the optimization.

            ["nl_means", h_out, patch_size_out, patch_dist_out]

                h_out: Cut-off distance (in gray levels). The higher
                    h_out, the more permissive one is in accepting
                    patches. A higher h_out results in a smoother
                    image, at the expense of blurring features. For a
                    Gaussian noise of standard deviation sigma, a rule
                    of thumb is to choose the value of h_out to be
                    on the same order of magnitude as the standard 
                    deviation of the Gaussian noise. This is a float.

                patch_size_out: Size, in pixels, of the patches used for
                    denoising. Should be an integer.

                patch_dist_out: Maximal distance, in pixels, where to 
                    search for patches used for denoising.
            
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image after performing the
        image processing procedures. img_out is in the same format as
        img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    if fltr_name == "nl_means":
        h_in = fltr_params[1]
        patch_size_in = fltr_params[2]
        patch_dist_in = fltr_params[3]

        sig_noise = rest.estimate_sigma(img_as_float(img), 
            average_sigmas=True, channel_axis=None)

        img_fltr = rest.denoise_nl_means(img, h=h_in, 
            sigma=sig_noise, fast_mode=True, patch_size=patch_size_in, 
            patch_distance=patch_dist_in, channel_axis=None)

        img_fltr = img_as_ubyte(img_fltr)
        
        if not quiet:
            print(f"\nSuccessfully applied the 'nl_means' denoising filter:\n"\
                f"    'h' (cut-off range in gray levels): {h_in}\n"\
                f"    Patch size: {patch_size_in}\n"\
                f"    Maximum search distance: {patch_dist_in}" )

    elif fltr_name == "tv_chambolle":
        weight_in = fltr_params[1]
        eps_in = fltr_params[2]
        n_iter_max_in = fltr_params[3]

        img_fltr = rest.denoise_tv_chambolle(img, weight=weight_in,
            eps=eps_in, n_iter_max=n_iter_max_in, multichannel=False, 
            channel_axis=None)

        img_fltr = img_as_ubyte(img_fltr)

        if not quiet:
            print(f"\nSuccessfully applied the 'tv_chambolle' denoising filter:\n"\
                f"    Denoising weight: {weight_in}\n"\
                f"    EPS stop criterion: {eps_in}\n"\
                f"    Maximal number of iterations: {n_iter_max_in}" )

    else:
        print(f"\nERROR: {fltr_name} is not a supported denoise filter.")
        img_fltr = img

    return img_fltr


def interact_driver_ridge_filter(img_in, fltr_name_in, flood_ext_in=True):
    """
    Interactively implements a "ridge" filter. This filter can be used
    to detect continuous ridges, e.g. tubes, wrinkles, rivers. This
    type of filter is also referred to as a ridge operator. The output
    parameters of this function are suitable for input for
    apply_driver_ridge_filter() below.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_name_in: A string that represents the name of the ridge
        filter to be applied. Currently, only "sato_tubeness" is
        supported.  
            
            "sato_tubeness": Filter an image with the Sato tubeness 
                filter. Calculates the eigenvectors of the Hessian to
                compute the similarity of an image region to tubes,
                according to the method described in Sato et al. (1998)
                "Three-dimensional multi-scale line filter for 
                segmentation and visualization of curvilinear structures
                in medical images" DOI:10.1016/S1361-8415(98)80009-1

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same format as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Local Copies ----
    fltr_name = fltr_name_in.lower()
    img_0 = img_in.copy()

    # List of strings of supported filter types
    fltr_list = ["sato_tubeness"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported filter type. Supported "\
            f"filter types are: \n  {fltr_list}"\
            "\nDefaulting to Sato tubeness.")
        fltr_name = "sato_tubeness"

    # Apply a flood-fill operation to the exterior region
    if flood_ext_in:
        flood_val = np.mean(img_0[img_0>0])
        flood_val = (np.round(flood_val)).astype(np.uint16)

        img_temp = np.pad(img_0, 1, mode='constant', 
            constant_values=img_0[0,0])

        img_temp = img_as_ubyte(img_temp)

        img_temp = seg.flood_fill(img_temp, (0,0), flood_val, 
            connectivity=1)

        img_0 = img_temp[1:-1, 1:-1]

    # -------- Filter Type: "sato_tubeness" --------
    if fltr_name == "sato_tubeness":
        [img_fltr, fltr_params] = ifun.interact_sato_tubeness(img_0)

    # Using this function just to write to standard output
    apply_driver_ridge_filter(img_0, fltr_params)

    return img_fltr


def apply_driver_ridge_filter(img_in, fltr_params_in, flood_ext_in=True,
    quiet_in=False):
    """
    Applies a "ridge" filter. This filter can be used to detect
    continuous ridges, e.g. tubes, wrinkles, rivers. This type of
    filter is also referred to as a ridge operator. This is the
    non-interactive version of interact_driver_ridge_filter().

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.
    
    fltr_params_in: A list of parameters needed to perform the 
        ridge filter. The first parameter is a string, which
        determines what type of ridge filter to be applied, as well
        as the definitions of the remaining parameters. As of now,
        fltr_params_in[0] must be "sato_tubeness". 
        Example parameter lists are given below for each type, 
            
            ["sato_tubeness", mask_val_out, sig_max_out, blk_ridges_out]

                mask_val_out: Only values greater than mask_val_out will
                    be altered by this filter. Hence, this integer acts
                    as a basic mask.

                sig_max_out: The maximum sigma used to scale  the 
                    filter. A range of sigma values will automatically
                    creates from one to sig_max_out in steps of one,
                    which is then utilized by the Sato filter. See the
                    Skimage documentation for more details. Typically
                    a value of ten is appropriate.

                blk_ridges_out: A boolean that affects whether the 
                    filter should detect black ridges or white ridges.
                    When True, the filter detects black ridges.
            
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image after performing the
        image processing procedures. img_out is in the same format as
        img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    fltr_name = (fltr_params[0]).lower()

    # Apply a flood-fill operation to the exterior region
    if flood_ext_in:
        flood_val = np.mean(img[img>0])
        flood_val = (np.round(flood_val)).astype(np.uint16)

        img_temp = np.pad(img, 1, mode='constant', 
            constant_values=img[0,0])

        img_temp = img_as_ubyte(img_temp)

        img_temp = seg.flood_fill(img_temp, (0,0), flood_val, 
            connectivity=1)

        img = img_temp[1:-1, 1:-1]

    if fltr_name == "sato_tubeness":
        mask_val_in = fltr_params[1]
        sig_max_in = fltr_params[2]
        blk_ridges_in = fltr_params[3]

        mask = (img > mask_val_in)

        img_temp = filt.sato(img, sigmas=range(1, sig_max_in), 
            black_ridges=blk_ridges_in)*mask

        img_temp = img_temp/np.amax(img_temp)
        img_fltr = img_as_ubyte(img_temp)

        if not quiet:
            print(f"\nSuccessfully applied the 'sato_tubeness' denoising filter:\n"\
                f"    Lower masking limit of grayscale values: {mask_val_in}\n"\
                f"    Max sigma used as a scale: {sig_max_in}\n"\
                f"    Detect black ridges (boolean): {blk_ridges_in}" )

    else:
        print(f"\nERROR: {fltr_name} is not a supported ridge filter.")
        img_fltr = img

    return img_fltr


def interact_driver_morph(img_in):
    """
    Interactively applies a morphological operation on a binary
    (i.e., segmented) image. More specifically, apply either an
    erosion, dilation, "opening", or "closing" with the option of
    choosing different kernel shapes and sizes. The implementations of
    these operations are based on the Skimage library. The output
    parameters of this function are suitable for input for
    apply_driver_morph_3d() below.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a grayscale image. It is assumed that the
        image is already grayscale and of type uint8. The array should
        thus be 2D, where each value represents the intensity for each
        corresponding pixel.

    ---- RETURNED ----
    [img_out]: Returns the final image after closing the interactive
        session. img_out is in the same format as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Local Copies ----
    img_0 = img_in.copy()

    # Run the interactive morphological operations GUI
    [img_fltr, fltr_params] = ifun.interact_binary_morph(img_0)

    # Run the driver application simply for the text output
    apply_driver_morph(img_0, fltr_params)

    return img_fltr


def apply_driver_morph(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a morphological operation on a binary (i.e., segmented)
    image. More specifically, apply either an erosion,
    dilation, "opening", or "closing" with the option of choosing
    different kernel shapes and sizes. The implementations of these
    operations are based on the Skimage library. This is the 2D,
    non-interactive version of interact_driver_morph(). See 
    apply_driver_morph_3d() for a 3D version of this function.

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a 2D grayscale image. It is 
        assumed that the image is already grayscale and of type uint8.
        Each value in the 2D array represents the intensity for each
        corresponding pixel.
    
    fltr_params_in: A list of parameters needed to perform the 
        morphological operation. The three input parameters for this
        list are described below, 
            
            [operation_type_out, footprint_type_out, n_radius_out]

                operation_type_out: An integer flag that determines what 
                    type of morphological operation to perform:

                        0: binary_closing
                        1: binary_opening
                        2: binary_dilation
                        3: binary_erosion

                footprint_type_out: An integer flag that determines what  
                    type of 2D neighborhood to use:

                        0: square 
                        1: disk 
                        2: diamond 

                n_radius_out: Radius of the footprint neighborhood in 
                    pixels, as an integer.
            
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image array after performing
        the image processing procedures. img_out is in the same format 
        as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img_0 = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    # Determine what type of morphological operation to perform
    # 0: binary_closing
    # 1: binary_opening
    # 2: binary_dilation
    # 3: binary_erosion
    operation_type = fltr_params[0]

    # Determine what type of 2D neighborhood to use.
    # 0: square (which corresponds to a cube in 3D)
    # 1: disk (which corresponds to a ball in 3D)
    # 2: diamond (which corresponds to a octahedron in 3D)
    footprint_type = fltr_params[1]

    # Radius of the footprint neighborhood (in pixels)
    n_radius = fltr_params[2]

    if not quiet:
        print(f"\nApplying morphological operation...")

    if footprint_type == 0:
        temp_width = 2*n_radius + 1
        temp_footprint = morph.square(temp_width)
        
        if not quiet:
            print(f"    Footprint type: 'square'")
            print(f"    Footprint radius (pixels): {n_radius}")

    elif footprint_type == 1:
        temp_footprint = morph.disk(n_radius)

        if not quiet:
            print(f"    Footprint type: 'disk'")
            print(f"    Footprint radius (pixels): {n_radius}")

    else: # footprint_type == 2
        temp_footprint = morph.diamond(n_radius)

        if not quiet:
            print(f"    Footprint type: 'diamond'")
            print(f"    Footprint radius (pixels): {n_radius}")

    if operation_type == 0:
        img_temp = morph.binary_closing(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_closing'")

    elif operation_type == 1:
        img_temp = morph.binary_opening(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_opening'")

    elif operation_type == 2:
        img_temp = morph.binary_dilation(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_dilation'")

    else: # operation_type == 3
        img_temp = morph.binary_erosion(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_erosion'")

    img_fltr = img_as_ubyte(img_temp)

    return img_fltr


def apply_driver_morph_3d(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a morphological operation on a binary (i.e., segmented)
    image. More specifically, apply either an erosion,
    dilation, "opening", or "closing" with the option of choosing
    different kernel shapes and sizes. The implementations of these
    operations are based on the Skimage library. This is the 3D,
    non-interactive version of interact_driver_morph().

    ---- INPUT ARGUMENTS ----
    [img_in]: Numpy array for a 3D set of grayscale images. It is 
        assumed that the images are already grayscale and of type 
        uint8. Each value in the 3D array represents the intensity 
        for each corresponding pixel.
    
    fltr_params_in: A list of parameters needed to perform the 
        morphological operation. The three input parameters for this
        list are described below, 
            
            [operation_type_out, footprint_type_out, n_radius_out]

                operation_type_out: An integer flag that determines what 
                    type of morphological operation to perform:

                        0: binary_closing
                        1: binary_opening
                        2: binary_dilation
                        3: binary_erosion

                footprint_type_out: An integer flag that determines what  
                    type of 3D neighborhood to use:

                        0: cube 
                        1: ball 
                        2: octahedron 

                n_radius_out: Radius of the footprint neighborhood in 
                    pixels, as an integer.
            
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.

    ---- RETURNED ----
    [img_out]: Returns the resultant Numpy image array after performing
        the image processing procedures. img_out is in the same format 
        as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. Strings are 
    printed to standard output.
    """

    # ---- Local Copies ----
    img_0 = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in

    # Determine what type of morphological operation to perform
    # 0: binary_closing
    # 1: binary_opening
    # 2: binary_dilation
    # 3: binary_erosion
    operation_type = fltr_params[0]

    # Determine what type of 3D neighborhood to use.
    # 0: cube
    # 1: ball
    # 2: octahedron
    footprint_type = fltr_params[1]

    # Radius of the footprint neighborhood (in pixels)
    n_radius = fltr_params[2]

    if not quiet:
        print(f"\nApplying morphological operation...")

    if footprint_type == 0:
        temp_width = 2*n_radius + 1
        temp_footprint = morph.cube(temp_width)
        
        if not quiet:
            print(f"    Footprint type: 'cube'")
            print(f"    Footprint radius (pixels): {n_radius}")

    elif footprint_type == 1:
        temp_footprint = morph.ball(n_radius)

        if not quiet:
            print(f"    Footprint type: 'ball'")
            print(f"    Footprint radius (pixels): {n_radius}")

    else: # footprint_type == 2
        temp_footprint = morph.octahedron(n_radius)

        if not quiet:
            print(f"    Footprint type: 'octahedron'")
            print(f"    Footprint radius (pixels): {n_radius}")

    if operation_type == 0:
        img_temp = morph.binary_closing(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_closing'")

    elif operation_type == 1:
        img_temp = morph.binary_opening(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_opening'")

    elif operation_type == 2:
        img_temp = morph.binary_dilation(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_dilation'")

    else: # operation_type == 3
        img_temp = morph.binary_erosion(img_0, footprint=temp_footprint)

        if not quiet:
            print(f"    Operation type: 'binary_erosion'")

    img_fltr = img_as_ubyte(img_temp)

    return img_fltr



from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy


# To compile the Cython extensions, run the following Python command
# in the terminal, which will execute this setup file. You must, of course,
# be in the same directory as this file.
#
# python cython_build_setup.py build_ext --build-lib ./bin
#
# This will build the Python/Cython source files (with .pyx extensions)
# in "../functions/cython" and translate them into C files within the same
# directory. The resultant binaries (with .pyd extensions in Windows) will 
# be in "../functions/cython/bin". Note, the files will be compiled in the
# scratch directory "./build".


extensions = [
    Extension("im3_processing", 
        ["./im3_processing.pyx"]
    )
]

setup(
    name="XCT Cython Extensions",

    ext_modules = cythonize(
        extensions,
        compiler_directives={'language_level' : "3"}
    ),

    include_dirs=[numpy.get_include()]
)


# Import external dependencies
import numpy as np
import matplotlib.pyplot as plt

from skimage.util import img_as_ubyte


def create_bw_fig(img_in, show_histogram=False):
    """
    Creates a black and white figure. The inputs are an 8-bit image,
    defined as a 2D Numpy array, and a boolean. When the boolean is set
    to True, a histogram of the image will also be shown. 
    """

    img = img_in.copy()
    #img = img_as_ubyte(img)

    if show_histogram:
        # (Width, Height) in inches
        fig_size = (12,6)

        # Show the histogram to the right of the grayscale image
        fig1, ax1 = plt.subplots(1, 2)

        # Big enough, and still fit in modern monitors
        fig1.set_size_inches(fig_size[0], fig_size[1])

        # Include vmin & vmax, else imshow() automatically normalizes from 0 to 1
        ax1[0].set_aspect('equal')
        ax1[0].imshow(img, cmap='gray', vmin=0, vmax=255)
        #ax1[0].set_title("Image")
        ax1[0].set_xlabel("X Pixel Number")
        ax1[0].set_ylabel("Y Pixel Number")

        ax1[1].hist(img.ravel(),256,[0,256])
        #ax1[1].set_title("Histogram")
        ax1[1].set_xlabel("Grayscale Intensity")
        ax1[1].set_ylabel("Counts") 


    else:
        # (Width, Height) in inches
        fig_size = (6,6)

        # Using subplot to make a single image since it conveniently returns
        # the image figure and axes objects
        fig1, ax1 = plt.subplots(1, 1)

        # Big enough, and still fit in modern monitors
        fig1.set_size_inches(fig_size[0], fig_size[1])

        # Include vmin & vmax, else imshow() automatically normalizes from 0 to 1
        ax1.set_aspect('equal')
        ax1.imshow(img, cmap='gray', vmin=0, vmax=255)
        #ax1.set_title("Image")
        ax1.set_xlabel("X Pixel Number")
        ax1.set_ylabel("Y Pixel Number") 


    # This function returns the figure object (as fig1) and the axes object
    # (as ax1). To actually show these images in the calling script, type the
    # following:

    """
        # Makes the subplots fit better in the figure's canvas. 
        plt.tight_layout() 

        plt.show()
    """

    # Note, this function is blocking: the script will not continue until
    # the figures are closed. To show the figures in a non-blocking manner,
    # use the following code:

    """
        # Makes the subplots fit better in the figure's canvas. 
        plt.tight_layout() 

        plt.show(block=False)
        plt.pause(1.0) # Pause for one second while the figure loads
    """

    # In this case, it is up to the user to ensure that the figure reference
    # is closed before the script terminates. To do so, use the plt.close()
    # function, i.e.  plt.close('all'). 

    return [fig1, ax1]


def create_2_bw_figs(img1_in, img2_in):
    """
    Creates 2 black and white figures, side-by-side. The inputs are two
    8-bit images, each defined as a 2D Numpy array. 
    """

    img_1 = img1_in.copy()
    img_2 = img2_in.copy()
    #img_1 = img_as_ubyte(img_1)
    #img_2 = img_as_ubyte(img_2)

    # (Width, Height) in inches
    fig_size = (12,6)

    # Link the axes of both images
    fig1, ax1 = plt.subplots(1, 2, sharex='row', sharey='row')

    # Big enough, and still fit in modern monitors
    fig1.set_size_inches(fig_size[0], fig_size[1])

    # Include vmin & vmax, else imshow() automatically normalizes from 0 to 1
    ax1[0].set_aspect('equal')
    ax1[0].imshow(img_1, cmap='gray', vmin=0, vmax=255)
    #ax1[0].set_title("Image 1")
    ax1[0].set_xlabel("X Pixel Number")
    ax1[0].set_ylabel("Y Pixel Number")

    ax1[1].set_aspect('equal')
    ax1[1].imshow(img_2, cmap='gray', vmin=0, vmax=255)
    #ax1[1].set_title("Image 2")
    ax1[1].set_xlabel("X Pixel Number")
    ax1[1].set_ylabel("Y Pixel Number")

    # This function returns the figure object (as fig1) and the axes object
    # (as ax1). To actually show these images in the calling script, type the
    # following:

    """
        # Makes the subplots fit better in the figure's canvas. 
        plt.tight_layout() 

        plt.show()
    """

    # Note, this function is blocking: the script will not continue until
    # the figures are closed. To show the figures in a non-blocking manner,
    # use the following code:

    """
        # Makes the subplots fit better in the figure's canvas. 
        plt.tight_layout() 

        plt.show(block=False)
        plt.pause(1.0) # Pause for one second while the figure loads
    """

    # In this case, it is up to the user to ensure that the figure reference
    # is closed before the script terminates. To do so, use the plt.close()
    # function, i.e.  plt.close('all').

    return [fig1, ax1]






# Import external dependencies
import numpy as np
import cv2 as cv

# Import local packages
from . import cv_processing_wrappers as wrap
from . import cv_interactive_processing as ifun


def interact_driver_blur(img_in, fltr_name_in):
    """
    Interactively implements a blur filter on the provided image. The
    parameters for the image processing are controlled by trackbars in
    the interactive window, and the resultant effects are updated in
    real-time. Upon hitting the Enter or Esc keys, the interactive 
    window closes, and then this function returns the final image. Basic
    details of the supported blur filters are given below, but more
    details are available at "https://docs.opencv.org/4.2.0/d4/d13/
    tutorial_py_filtering.html". This function is related to 
    apply_driver_blur(...).

    ---- INPUT ARGUMENTS ----
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.
    
    fltr_name_in: A string that represents the name of the blur filter
        to be applied. Can be either "average", "gaussian", "median", or
        "bilateral".  
            
            "average": Equal-weighted averaging kernel
            
            "gaussian": Gaussian-weighted kernel
            
            "median": The center weight of kernal is the median value
            
            "bilateral": Edge-preserving Gaussian kernel

    ---- RETURNED ----
    [img_fltr]: Returns the final image after closing the interactive
        session. img_fltr is in the same format as img_in.

    ---- SIDE EFFECTS ----
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    fltr_name = fltr_name_in.lower() # Ensure string is all lowercase
    # ---- End Local Copies ----

    # List of strings of the supported filter types
    fltr_list = ["average", "gaussian", "median", "bilateral"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported filter type. Supported "\
            f"filter types are: {fltr_list}\nDefaulting to a Gaussian.")
        fltr_name = "gaussian"

    # -------- Filter Type: "average" --------
    if fltr_name == "average":
        [img_fltr, fltr_params] = ifun.interact_average_filter(img_in)

    # -------- Filter Type: "gaussian" --------
    elif fltr_name == "gaussian":
        [img_fltr, fltr_params] = ifun.interact_gaussian_filter(img_in)

    # -------- Filter Type: "median" --------
    elif fltr_name == "median":
        [img_fltr, fltr_params] = ifun.interact_median_filter(img_in)

    # -------- Filter Type: "bilateral" --------
    elif fltr_name == "bilateral":
        [img_fltr, fltr_params] = ifun.interact_bilateral_filter(img_in)
    
    # Using this function just to write to standard output
    apply_driver_blur(img_in, fltr_params)

    return img_fltr


def apply_driver_blur(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a blur filter to the provided image based on the input
    parameters. This is the non-interactive version of the related
    function, interact_driver_blur(...). For more details about the
    supported blur filters, go to "https://docs.opencv.org/4.2.0/d4/
    d13/tutorial_py_filtering.html".

    ---- INPUT ARGUMENTS ----
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.
    
    [fltr_params_in]: A list of parameters needed to perform the blur
        operation. The first parameter is a string, which determines
        what type of blur filter to be applied, as well as the 
        definitions of the remaining parameters. fltr_params_in[0] can
        either be "average", "gaussian", "median", or "bilateral". 
        Example parameter lists are given below for each type,
            
            ["average", (k_size, k_size)]
                k_size: Kernel size for the average filter. If zero, 
                    then no filter was applied.
            
            ["gaussian", (k_size,k_size), std_dev]
                k_size: Kernel size for the Gaussian filter. If zero, 
                    then no filter was applied. Must be odd.
                std_dev: Standard deviation for the Gaussian kernel. If
                    std_dev < 0, then the standard deviation is
                    automatically calucated using, 
                    0.3*((k_size - 1)*0.5 - 1) + 0.8
            
            ["median", k_size]
                k_size: Kernel size for the median filter. If zero, 
                    then no filter was applied. Must be odd.
            
            ["bilateral", d_size, sig_intsty]
                d_size: Diameter of each pixel neighborhood (must be 
                    even). If zero, then no filter was applied.
                sig_intsty: Filter sigma in the color space. A larger
                    value means that farther colors within d_size will 
                    be mixed together, resulting in larger areas of
                    semi-equal color. 
    
    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing.      

    ---- RETURNED ---- 
    [img_fltr]: Returns the resultant image after performing the
        image processing procedures. img_fltr is in the same format as
        img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output. 
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in
    # ---- End Local Copies ----

    fltr_name = (fltr_params[0]).lower()

    # Extract the filter parameters depending on the filter type, and
    # then apply said filter. Note, it is possible that some parameters,
    # like kernel size, are zero which would throw an error if used in
    # some of OpenCV's filters. If kernel size is zero, then that means
    # just show the original image -- no changes were made.
    if fltr_name == "average": # Equal-weighted average kernel
        avg_ksize = fltr_params[1]
        if avg_ksize == (0, 0): # No filtering performed, keep original
            return img 
        else: # Apply filter
            img_fltr = cv.blur(img, avg_ksize)

        if not quiet:
             print(f"\nSuccessfully appled the 'average' blur filter:\n"\
                f"    Kernel Size = {(avg_ksize, avg_ksize)}")

    elif fltr_name == "gaussian": # Gaussian-weighted kernel
        gaus_ksize = fltr_params[1]
        gaus_sdev = fltr_params[2]
        if gaus_ksize == (0, 0): # No filtering performed, keep original
            return img 
        else: # Apply filter
            if gaus_sdev < 0: # Calculate automatically if less than 0
                gaus_sdev = 0.3*((gaus_ksize - 1)*0.5 - 1) + 0.8

            if gaus_ksize[0] % 2 == 0: # Must be odd for this filter
                gaus_ksize[0] += -1
                gaus_ksize[1] += -1

            img_fltr = cv.GaussianBlur(img, gaus_ksize, gaus_sdev)

        if not quiet:
            print(f"\nSuccessfully applied the 'gaussian' blur filter:\n"\
                f"    Kernel Size = {(gaus_ksize, gaus_ksize)}\n"\
                f"    Standard Deviation = {gaus_sdev}") 

    elif fltr_name == "median": # Center weight of kernal is median value
        med_ksize = fltr_params[1]
        if med_ksize == 0: # No filtering performed, keep original
            return img 
        else: # Apply filter
            if med_ksize % 2 == 0: # Must be odd for this filter
                med_ksize += -1
            img_fltr = cv.medianBlur(img, med_ksize) 

        if not quiet:
            print(f"\nSuccessfully applied the 'median' blur filter:\n"\
                f"    Kernel Size = {med_ksize}")

    elif fltr_name == "bilateral": # Edge-preserving Gaussian kernel
        bil_dsize = fltr_params[1]
        bil_sint = fltr_params[2]
        if bil_dsize == 0: # No filtering performed, keep original
            return img 
        else: # Apply filter
            if bil_dsize % 2 != 0: # Should be even for this filter
                bil_dsize += 1
            img_fltr = cv.bilateralFilter(img, bil_dsize, bil_sint, bil_sint)

        if not quiet:
            print(f"\nSuccessfully applied the 'bilateral' blur filter:\n"\
                f"    Pixel Neighborhood = {bil_dsize}\n" \
                f"    Intensity Threshold = {bil_sint}")

    else:
        print(f"\nERROR: {fltr_name} is not a supported blur filter.")
        img_fltr = img

    return img_fltr


def interact_driver_sharpen(img_in, fltr_name_in):
    """
    Interactively implements a sharpen filter on the provided image. The
    parameters for the image processing are controlled by trackbars in
    the interactive window, and the resultant effects are updated in
    real-time. Upon hitting the Enter or Esc keys, the interactive 
    window closes, and then this function returns the final image. These
    image processing operations can also be called edge enhancements.
    This function is related to apply_driver_sharpen(...).

    ---- INPUT ARGUMENTS ----
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    fltr_name_in: A string that represents the name of the sharpen
        filter to be applied. Can be either "unsharp", "laplacian", or
        "canny", 
            
            "unsharp": Applies a conventional unsharp mask based on 
                either a Gaussian or median kernel. See digital unsharp 
                masking at, "https://en.wikipedia.org/wiki/
                Unsharp_masking"
            
            "laplacian": Uses the Laplacian (i.e., second derivatives)
                to enhance the image's edges.
            
            "canny": Uses the Canny edge detection algorithm to add
                contrast to the edges. These enhancements may be 
                somewhat discrete, rather than smooth gradients.

    ---- RETURNED ---- 
    [img_sharp]: Returns the final image after closing the interactive
        session. img_sharp is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    fltr_name = fltr_name_in.lower() # Ensure string is all lowercase
    # ---- End Local Copies ----

    # List of strings of the supported filter types
    fltr_list = ["unsharp", "laplacian", "canny"]

    # Ensure that the filter name is recognized
    if fltr_name not in fltr_list:
        print(f"\n{fltr_name_in} is not a supported filter type. Supported "\
            f"filter types are: {fltr_list}\nDefaulting to an Unsharp Mask.")
        fltr_name = "unsharp"

    # -------- Filter Type: "unsharp" --------
    if fltr_name == "unsharp":
        [img_sharp, sharp_params] = ifun.interact_unsharp_mask(img_in)

    # -------- Filter Type: "laplacian" --------
    elif fltr_name == "laplacian":
        [img_sharp, sharp_params] = ifun.interact_laplacian_sharp(img_in)

    # -------- Filter Type: "canny" --------
    elif fltr_name == "canny":
        [img_sharp, sharp_params] = ifun.interact_canny_sharp(img_in)

    # Using this function just to write to standard output
    apply_driver_sharpen(img_in, sharp_params)

    return img_sharp


def apply_driver_sharpen(img_in, fltr_params_in, quiet_in=False):
    """
    Applies a sharpen filter, much like interact_driver_sharpen(...),
    but in a non-interactive way. Note, sharpening is effectively the 
    same as edge-enhancement.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    [fltr_params_in]: A list of parameters needed to perform the sharpen
        operation. The first parameter is a string, which determines
        what type of sharpen filter to be applied, as well as the 
        definitions of the remaining parameters. fltr_params_in[0] can
        either be "unsharp", "laplacian", "canny. Example parameter 
        lists are given below for each type,
            
            ["unsharp", cur_amount, k_size, fltr_type, std_dev]
                cur_amount: A percentage value of how much to blend
                    the sharpening effect onto the original value. It
                    corresponds to (amount)*100 in the equation above.
                    If zero, then no filter is applied.
                k_size: Kernel size for the blur filter. If zero, 
                    then no filter was applied. Must be odd.
                fltr_type: Either 0 or 1. If 0, then the blurred image
                    used in the procedure utilizes the Gaussian filter.
                    If 1, then the median filter is used.
                std_dev: Corresponds to the standard deviation used in
                    defining the Gaussian kernel for the Gaussian
                    filter. This parameter is ignored if the median
                    filter is used. If std_dev < 0, then the standard
                    deviation is automatically calculated.
            
            ["laplacian", cur_amount, blur_k_size, lap_k_size,  
            fltr_type, blur_std_dev]
                cur_amount: A percentage value of how much to blend
                    the sharpening effect onto the original value. It
                    corresponds to (amount)*100 in the equation above.
                    If zero, then no filter is applied.
                blur_k_size: Kernel size for the blur filter that is
                    applied prior to using the Laplacian kernel. If 
                    zero, then no blur filter is applied. Must be odd.
                lap_k_size: Kernel size for the Laplacian kernel used
                    to find the edges of the original image.
                fltr_type: Either 0 or 1. If 0, then the blurred image
                    used in the procedure utilizes the Gaussian filter.
                    If 1, then the median filter is used.
                blur_std_dev: Corresponds to the standard deviation used
                    in defining the Gaussian kernel for the Gaussian
                    blur. This parameter is ignored if the median blur
                    is used. If std_dev < 0, then the standard deviation
                    is automatically calculated.
            
            ["canny", k_size, cur_amount, thresh1, thresh2]
                k_size: Kernel size for the Canny edge algorithm. If
                    zero, then no filter was applied. Must be odd.
                cur_amount: A percentage value of how much to blend
                    the sharpening effect onto the original value. It
                    corresponds to (amount)*100 in the equation above.
                    If zero, then no filter is applied.
                thresh1: Related to the hystersis thresholding algorithm
                    in the Canny Edge Detection procedure. Related to 
                    minimum threshold used to determine what an edge is.
                thresh2: Related to the hystersis thresholding algorithm
                    in the Canny Edge Detection procedure. Related to 
                    maximum threshold used to determine what an edge is.
    
    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing. 

    ---- RETURNED ---- 
    [img_sharp]: Returns the resultant image after performing the
        image processing procedures. img_sharp is in the same format as
        img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    fltr_params = fltr_params_in
    quiet = quiet_in
    # ---- End Local Copies ----

    fltr_name = (fltr_params[0]).lower()

    # -------- Filter Type: "unsharp" --------
    if fltr_name == "unsharp":
        fltr_params_2 = fltr_params[1:]

        if fltr_params_2[1] % 2 == 0: # Must be odd for the blur filters
            fltr_params_2[1] += -1

        if fltr_params_2[0] <= 0:
            return img
        elif fltr_params_2[1] < 1:
            return img

        img_sharp = wrap.unsharp_mask(img, fltr_params_2)

        if not quiet:
            print(f"\nSuccessfully applied the 'unsharp' mask:\n"\
                f"    Amount: {fltr_params_2[0]}%\n"\
                f"    Radius of Blur Kernel: {fltr_params_2[1]}")

            if fltr_params_2[2]:
                print("    Blur Filter Type: Median")
            else:
                print("    Blur Filter Type: Gaussian\n"\
                    f"    Standard Deviation: {fltr_params_2[3]}")

    # -------- Filter Type: "laplacian" --------
    elif fltr_name == "laplacian":
        fltr_params_2 = fltr_params[1:]

        if fltr_params_2[1] % 2 == 0: # Must be odd for the Gaussian blur
            fltr_params_2[1] += -1

        if fltr_params_2[2] % 2 == 0: 
            fltr_params_2[2] += -1

        if fltr_params_2[0] <= 0:
            return img
        elif (fltr_params_2[2] < 1) or (fltr_params_2[1] < 1):
            return img

        img_sharp = wrap.laplacian_sharp(img, fltr_params_2)

        if not quiet:
            print(f"\nSuccessfully applied the 'laplacian' mask:\n"\
                f"    Amount: {fltr_params_2[0]}%\n"\
                f"    Radius of Blur Kernel: {fltr_params_2[1]}\n"\
                f"    Radius of Laplacian Kernel: {fltr_params_2[2]}")

            if fltr_params_2[3]:
                print("    Blur Filter Type: Median")
            else:
                print("    Blur Filter Type: Gaussian\n"\
                    f"    Standard Deviation: {fltr_params_2[4]}")

    # -------- Filter Type: "canny" --------
    elif fltr_name == "canny":
        fltr_params_2 = fltr_params[1:]

        if fltr_params_2[0] % 2 == 0: 
            fltr_params_2[0] += -1

        img_sharp = wrap.canny_sharp(img, fltr_params_2)

        if not quiet:
            print(f"\nSuccessfully applied the 'canny' edge mask:\n"\
                f"    Radius of Canny Edge Kernel: {fltr_params_2[0]}\n"\
                f"    Amount: {fltr_params_2[1]}%\n"\
                f"    Hysteresis Threshold 1: {fltr_params_2[2]}\n"\
                f"    Hysteresis Threshold 2: {fltr_params_2[3]}")

    else:
        print(f"\nERROR: {fltr_name} is not a supported sharpen filter.")
        img_sharp = img
        
    return img_sharp


def interact_driver_equalize(img_in, eq_name_in):
    """
    Interactively equalizes the histogram on the provided image. The
    parameters for the image processing are controlled by trackbars in
    the interactive window, and the resultant effects are updated in
    real-time. Upon hitting the Enter or Esc keys, the interactive 
    window closes, and then this function returns the final image. These
    image processing operations can also be called contrast enhancements.
    This function is related to apply_driver_equalize(...). More details
    about equalization can be found at, "https://docs.opencv.org/4.2.0/
    d5/daf/tutorial_py_histogram_equalization.html"

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 
    
    eq_name_in: A string that represents the type of equalization 
        technique to use. Can be either "global" or "adaptive".
            
            "global": Warning, this is not interactive! Values are
                automatically calculated and applied to the entire 
                image.
            
            "adaptive": Locally adaptive procedure to enhance the
                contrast. This particular process is known as Contrast
                Limited Adaptive Histogram Equalization (CLAHE).

    ---- RETURNED ---- 
    [img_eq]: Returns the final image after closing the interactive
        session. img_eq is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    eq_name = eq_name_in.lower() # Ensure string is all lowercase
    # ---- End Local Copies ----

    # List of strings of the supported filter types
    eq_list = ["global", "adaptive"]

    # Ensure that the filter name is recognized
    if eq_name not in eq_list:
        print(f"\n{eq_name_in} is not a supported filter type. Supported "\
            f"filter types are: {eq_list}\nDefaulting to a Adaptive.")
        eq_name = "adaptive"

    # -------- Filter Type: "global" --------
    if eq_name == "global":
        img_eq = wrap.global_equalize(img_in)
        eq_params = ["global"]

    # -------- Filter Type: "adaptive" --------
    elif eq_name == "adaptive":
        [img_eq, eq_params] = ifun.interact_adaptive_equalize(img_in)

    # Using this function just to write to standard output
    apply_driver_equalize(img_in, eq_params)

    return img_eq


def apply_driver_equalize(img_in, eq_params_in, quiet_in=False):
    """
    Equalizes the image's intensity histogram, much like 
    interact_driver_equalize(...), but in a non-interactive way. Note,
    equalizing the histogram is a method to enhance contrast. More 
    details about equalization can be found at, "https://docs.opencv.org
    /4.2.0/d5/daf/tutorial_py_histogram_equalization.html"

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel.
    
    [eq_params_in]: A list of parameters needed to perform the 
        equalization operation. The first parameter is a string, which
        determines what type of equalization to be applied, as well as
        the  definitions of the remaining parameters. eq_params_in[0]
        can either be "global" or "adaptive". Example parameter lists
        are  given below for each type,
            
            ["global"]
                No additional parameters required
            
            ["adaptive", clip_limit, grid_size]
                clip_limit: Threshold for contrast limiting
                grid_size: Size of grid for histogram equalization. 
                    Input image will be divided into equally sized 
                    rectangular tiles. tileGridSize defines the number 
                    of tiles in row and column. If zero, the original
                    image is returned.]
    
    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing. 

    ---- RETURNED ---- 
    [img_eq]: Returns the resultant image after performing the
        image processing procedures. img_eq is in the same format as
        img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    eq_params = eq_params_in
    quiet = quiet_in
    # ---- End Local Copies ----

    eq_name = (eq_params[0]).lower()

    # -------- Filter Type: "global" --------
    if eq_name == "global":
        img_eq = wrap.global_equalize(img)
        if not quiet:
            print("\nSuccessfully applied 'global' equalization")

    # -------- Filter Type: "adaptive" --------
    elif eq_name == "adaptive":
        eq_params_2 = eq_params[1:]

        if eq_params_2[1] < 1:
            return img

        img_eq = wrap.adaptive_equalize(img, eq_params_2)

        if not quiet:
            print("\nSuccessfully applied locally 'adaptive' equalization:\n"\
                f"    Clip Limit: {eq_params_2[0]}\n"\
                f"    Tile Size: ({eq_params_2[1]},{eq_params_2[1]})")

    else:
        print(f"\nERROR: {eq_name} is not a supported equalization "\
            "operation.")
        img_eq = img

    return img_eq


def interact_driver_morph(img_in):
    """
    Creates an interactive session to perform various types of 
    morphological transformations (i.e., dilation and erosion). The user
    can either perform just dilation operations, or just erosion
    transformations. Alternatively, the user can perform either an open
    transformation or a close transformation. Note, an open
    transformation is an erosion followed by a dilation, and it can be
    helpful in removing noise. Conversely, a close transformation is a
    dilation followed by an erosion, and it can be useful in closing
    small holes. The user can update the various parameters and see the
    effects in real-time within the interactive window. Upon hitting the
    Enter or Esc keys, the interactive window closes, and then this 
    function returns the final image. This function is closely related
    to apply_driver_morph(...).

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    ---- RETURNED ---- 
    [img_morph]: Returns the final image after closing the interactive
        session. img_morph is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    [img_morph, morph_params] = ifun.interact_morph(img_in)

    # Using this function just to write to standard output
    apply_driver_morph(img_in, morph_params)

    return img_morph


def apply_driver_morph(img_in, morph_params_in, quiet_in=False):
    """
    Applies a morphological operation (i.e., dilation and erosion
    transformations), much like interact_driver_morph(...) but not in an
    interactive way. The user can either perform just dilation
    operations, or just erosion transformations. Alternatively, the user
    can perform either an open transformation or a close transformation.
    Note, an open transformation is an erosion followed by a dilation,
    and it can be helpful in removing noise. Conversely, a close
    transformation is a dilation followed by an erosion, and it can be
    useful in closing small holes.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    [morph_params_in]: A list parameters needed to perform the desired
        morphological transformations. A more detailed explanation of
        each parameter in morph_params_in is given below,
            
            [flag_open_close, flag_rect_ellps, k_size, num_erode,
            num_dilate]
                flag_open_close: Either 0 or 1, and determines if the
                    erosion operations should occur prior to dilation
                    operations, or the other way around. If 0, then an 
                    open operation is done, which is erosion followed by 
                    dilation (removes noise). If 1, then a close 
                    operation is done, which is a dilation followed by
                    erosion (closes small holes). 
                flag_rect_ellps: Either 0 or 1. If 0, then a rectangular
                    kernel is used for the morphological operations. If
                    1, then an elliptical kernel is used.
                k_size: Kernel size for the morphological operations. If
                    less than 1, then the original image is returned.
                num_erode: Number of erosion operations to perform
                num_dilate: Number of dilation operations to perform
    
    quiet_in: A boolean that determines if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing. 

    ---- RETURNED ---- 
    [img_morph]: Returns the final image after the morphological 
        transformations. img_morph is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    morph_params = morph_params_in
    quiet = quiet_in

    # 0 is open => erosion followed by dilation (removes noise)
    # 1 is close => dilation followed by erosion (closes small holes)
    flag_open_close = morph_params[0]

    # 0 is rectangular kernel, 1 is elliptical kernel
    flag_rect_ellps = morph_params[1]

    # Width (or diameter) of the kernel in pixels
    k_size = morph_params[2]

    # Number of iterations of the erode and dilation operations
    num_erode = morph_params[3]
    num_dilate = morph_params[4]

    if k_size < 1: # Nothing to report if kept original image
        return img_in
    elif (num_dilate < 1) and (num_erode < 1):
        return img_in

    # Perform successive morphological operations
    img_morph = wrap.multi_morph(img_in, morph_params)

    # Write out comments to standard output based on what was done
    if flag_open_close: # Close
        if (num_dilate >= 1) and (num_erode >= 1):
            if not quiet:
                print("\nSuccessfully applied morphological operations:")
                if flag_rect_ellps:
                    print("    Kernel Shape: Ellipse\n"\
                         f"    Kernel Diameter: {k_size}")
                else:
                    print("    Kernel Shape: Rectangle\n"\
                         f"    Kernel Width: {k_size}")

                op_hist = f"{num_dilate} X Dilate   followed by   "\
                    f"{num_erode} X Erode"
                print("    Kernel Operations: " + op_hist)

        elif num_dilate < 1: # Only do a erosion
            if not quiet:
                print("\nSuccessfully applied morphological operations:")
                if flag_rect_ellps:
                    print("    Kernel Shape: Ellipse\n"\
                         f"    Kernel Diameter: {k_size}")
                else:
                    print("    Kernel Shape: Rectangle\n"\
                         f"    Kernel Width: {k_size}")

                op_hist = f"0 X Dilate   followed by   "\
                    f"{num_erode} X Erode"
                print("    Kernel Operations: " + op_hist)

        else: # Only do an dilation
            if not quiet:
                print("\nSuccessfully applied morphological operations:")
                if flag_rect_ellps:
                    print("    Kernel Shape: Ellipse\n"\
                         f"    Kernel Diameter: {k_size}")
                else:
                    print("    Kernel Shape: Rectangle\n"\
                         f"    Kernel Width: {k_size}")

                op_hist = f"{num_dilate} X Dilate   followed by   "\
                    f"0 X Erode"
                print("    Kernel Operations: " + op_hist)

    else: # Open
        if (num_dilate >= 1) and (num_erode >= 1):
            if not quiet:
                print("\nSuccessfully applied morphological operations:")
                if flag_rect_ellps:
                    print("    Kernel Shape: Ellipse\n"\
                         f"    Kernel Diameter: {k_size}")
                else:
                    print("    Kernel Shape: Rectangle\n"\
                         f"    Kernel Width: {k_size}")

                op_hist = f"{num_erode} X Erode   followed by   "\
                    f"{num_dilate} X Dilate"
                print("    Kernel Operations: " + op_hist)

        elif num_dilate < 1: # Only do a erosion
            if not quiet:
                print("\nSuccessfully applied morphological operations:")
                if flag_rect_ellps:
                    print("    Kernel Shape: Ellipse\n"\
                         f"    Kernel Diameter: {k_size}")
                else:
                    print("    Kernel Shape: Rectangle\n"\
                         f"    Kernel Width: {k_size}")

                op_hist = f"{num_erode} X Erode   followed by   "\
                    f"0 X Dilate"
                print("    Kernel Operations: " + op_hist)

        else: # Only do an dilation
            if not quiet:
                print("\nSuccessfully applied morphological operations:")
                if flag_rect_ellps:
                    print("    Kernel Shape: Ellipse\n"\
                         f"    Kernel Diameter: {k_size}")
                else:
                    print("    Kernel Shape: Rectangle\n"\
                         f"    Kernel Width: {k_size}")

                op_hist = f"0 X Erode   followed by   "\
                    f"{num_dilate} X Dilate"
                print("    Kernel Operations: " + op_hist)

    return img_morph


def interact_driver_thresh(img_in, thsh_name_in):
    """
    Creates an interactive session to perform a black and white
    threshold of a grayscale image. This is also known as binarization.
    Both global and locally adaptive methods are supported. The final
    image will be composed of just black and white intensities, which
    correspond to 0 and 255, respectively. Upon hitting the Enter or Esc
    keys, the interactive window closes, and then this  function returns
    the final image. This function is closely related to
    apply_driver_thresh(...).

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 
    
    thsh_name_in: A string that represents the name of the type of
        thresholding operation to be performed. Supported methods are
        "global", "adaptive_mean", and "adaptive_gaussian". Basic 
        descriptions are given below, but for more details, visit
        "https://docs.opencv.org/4.2.0/d7/d4d/tutorial_py_
        thresholding.html"

            "global": Uses the same threshold value (between 0 and 255)
                for every pixel. The threshold can be set manually or 
                automatically using Otsu's method.

            "adaptive_mean": A locally adaptive method. The threshold
                value is the mean of the neighborhood area minus a
                constant.

            "adaptive_gaussian": A locally adaptive method. The threshold
                value is a gaussian-weighted sum of the neighborhood area
                minus a constant.

    ---- RETURNED ---- 
    [img_thsh]: Returns the final image after closing the interactive
        session. img_thsh is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    thsh_name = thsh_name_in.lower()
    # ---- End Local Copies ----

    # List of strings of the supported filter types
    thsh_list = ["global", "adaptive_mean", "adaptive_gaussian"]

    # Ensure that the filter name is recognized
    if thsh_name not in thsh_list:
        print(f"\n{thsh_name} is not a supported threshold type. Supported "\
            f"threshold types are: {thsh_list}\nDefaulting to "\
            "adaptive_gaussian.")
        thsh_name = "adaptive_gaussian"

    if thsh_name == "global":
        [img_thsh, thsh_params] = ifun.interact_global_threshold(img_in)

    elif thsh_name == "adaptive_mean":
        [img_thsh, thsh_params] = ifun.interact_mean_threshold(img_in)

    elif thsh_name == "adaptive_gaussian":
        [img_thsh, thsh_params] = ifun.interact_gaussian_threshold(img_in)

    # Using this function just to write to standard output
    apply_driver_thresh(img_in, thsh_params)

    return img_thsh


def apply_driver_thresh(img_in, thsh_params_in, quiet_in=False):
    """
    Perform a black and white threshold of a grayscale image. This is
    also known as binarization. Both global and locally adaptive methods
    are supported. The final image will be composed of just black and
    white intensities, which correspond to 0 and 255, respectively. This
    function is similar to interact_driver_thresh(...), but is
    non-interactive.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    [thsh_params_in]: A list of parameters to perform the desired 
        threshold operation. The first parameter is a string, which
        determines what type of thresholding is performed, as well as
        the definitions of the remaining parameters. thsh_params_in[0]
        can be either "global", "adaptive_mean", or "adaptive_gaussian".
        Example parameter lists are given below for each type,

            ["global", cur_thsh]
                cur_thsh: The specified intensity to use as the 
                    threshold limit between black and white. It should
                    be an integer between 0 and 255. However, if
                    cur_thsh < 0, then cur_thsh will be calculated
                    automatically using Otsu's method.

            ["adaptive_mean", blk_size, c_offset]
                blk_size: Size of a pixel neighborhood that is used to
                    calculate a threshold value for the current pixel.
                    Must be odd, and if zero, no thresholding is done.
                c_offset: A constant offset value applied to the mean
                    or weighted mean of the intensities

            ["adaptive_gaussian", blk_size, c_offset]
                blk_size: Size of a pixel neighborhood that is used to
                    calculate a threshold value for the current pixel.
                    Must be odd, and if zero, no thresholding is done.
                c_offset: A constant offset value applied to the mean
                    or weighted mean of the intensities

    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing

    ---- RETURNED ---- 
    [img_thsh]: Returns the final image after thresholding. img_thsh is
        in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    # ---- Start Local Copies ----
    img = img_in.copy()
    thsh_params = thsh_params_in
    quiet = quiet_in
    # ---- End Local Copies ----

    thsh_name = (thsh_params[0]).lower()

    # -------- Filter Type: "global" --------
    if thsh_name == "global":
        thsh_params_2 = thsh_params[1:]

        # If the threshold value is < 0, apply automatic Otsu method
        if thsh_params_2[0] < 0:
            [ret_thsh, img_thsh] = cv.threshold(img, 0, 255, 
                cv.THRESH_BINARY+cv.THRESH_OTSU)

            if not quiet:
                print(f"\nSuccessfully applied 'global' thresholding:\n"\
                    f"    Threshold Value (Otsu Method): {ret_thsh}") 

        else:
            if thsh_params_2[0] > 255:
                thsh_params_2[0] = 255

            [ret_thsh, img_thsh] = cv.threshold(img, thsh_params_2[0], 255, 
                cv.THRESH_BINARY)

            if not quiet:
                print(f"\nSuccessfully applied 'global' thresholding:\n"\
                    f"    Threshold Value: {thsh_params_2[0]}")         

        return img_thsh


    # -------- Filter Type: "adaptive_mean" --------
    elif thsh_name == "adaptive_mean":
        thsh_params_2 = thsh_params[1:]

        if thsh_params_2[0] <= 0: # Return original image
            img_thsh = img

        else:
            if thsh_params_2[0] % 2 == 0: # Must be odd (but not 1) and > 0
                thsh_params_2[0] += -1

            if thsh_params_2[0] < 3:
                thsh_params_2[0] = 3

            img_thsh = cv.adaptiveThreshold(img, 255, 
                cv.ADAPTIVE_THRESH_MEAN_C, cv.THRESH_BINARY, 
                thsh_params_2[0], thsh_params_2[1])

            if not quiet:
                print(f"\nSuccessfully applied 'adaptive mean' threshold:\n"\
                    f"    Block Size: {thsh_params_2[0]} pixels\n"\
                    f"    Intensity Offset: {thsh_params_2[1]}")

        return img_thsh

    # -------- Filter Type: "adaptive_gaussian" --------
    elif thsh_name == "adaptive_gaussian":
        thsh_params_2 = thsh_params[1:]

        if thsh_params_2[0] <= 0: # Return original image
            img_thsh = img

        else:
            if thsh_params_2[0] % 2 == 0: # Must be odd (but not 1) and > 0
                thsh_params_2[0] += -1

            if thsh_params_2[0] < 3:
                thsh_params_2[0] = 3

            img_thsh = cv.adaptiveThreshold(img, 255, 
                cv.ADAPTIVE_THRESH_GAUSSIAN_C, cv.THRESH_BINARY, 
                thsh_params_2[0], thsh_params_2[1])

            if not quiet:
                print(f"\nSuccessfully applied 'adaptive gaussian' "\
                    f"threshold:\n"\
                    f"    Block Size: {thsh_params_2[0]} pixels\n"\
                    f"    Intensity Offset: {thsh_params_2[1]}")

        return img_thsh

    else:
        print(f"\nERROR: {thsh_name} is not a supported threshold "\
            "operation.")


def interact_driver_blob_fill(img_in):
    """
    Removes blobs (i.e., islands) of pixels from a binarized image in an
    interactive sesssion. Blobs are identified based on user-selected
    thresholding parameters related to the blob size (in pixels),
    circularity, and aspect ratio. It is assumed that the blobs to be
    removed are white, and thus, will be filled in with black pixels to
    effectively delete them. Upon hitting Enter during the interactive
    session, the interactive window closes and the final image is
    returned. This function is related to apply_driver_blob_fill(...).

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    ---- RETURNED ---- 
    [img_blob]: Returns the final image after closing the interactive
        session. img_blob is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """
    [img_blob, blob_params] = ifun.interact_blob_fill(img_in)

    # Using this function just to write to standard output
    apply_driver_blob_fill(img_in, blob_params)

    return img_blob


def apply_driver_blob_fill(img_in, blob_params_in, quiet_in=False):
    """
    Removes blobs (i.e., islands) of pixels from a binarized image like
    interact_driver_blob_fill(...), but as a batch process rather than in
    an interactive session. Blobs are identified based on user-selected
    thresholding parameters related to the blob size (in pixels),
    circularity, and aspect ratio. 

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed 
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    [blob_params_in]: A list of threshold parameters necessary to 
        identify blobs which should get filled in with a specified 
        color. More details on the parameters of blob_params_in is given
        below,

        [area_thresh_min, area_thresh_max, circty_thresh_min,
        circty_thresh_max, ar_min, ar_max, COLOR]
                
            area_thresh_min: Lower threshold bound for identifying blobs
                to be deleted. Related to the blob size based on the
                total number of pixels.
                
            area_thresh_max: Upper threshold bound for identifying blobs
                to be deleted. Related to the blob size based on the
                total number of pixels.
                
            circty_thresh_min: Lower threshold bound for identifying
                blobs to be deleted. Related to the circularity of the
                blob, based on, 4*pi*area/(perimeter)^2
                
            circty_thresh_max: Upper threshold bound for identifying
                blobs to be deleted. Related to the circularity of the
                blob, based on, 4*pi*area/(perimeter)^2
                
            ar_min: Lower threshold bound for identifying blobs to  be
                deleted. Related to the aspect ratio, which is the width
                over length of the bounding rectangle.
                
            ar_max: Upper threshold bound for identifying blobs to  be
                deleted. Related to the aspect ratio, which is the width
                over length of the bounding rectangle.
                
            COLOR: The color used to fill in the pixels of each blob.
                Should be a tuple containing three integers, each
                corresponding to the intensity of the color channels (in
                BGR format). For example, (0, 0, 0) corresponds to
                black.

    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing

    ---- RETURNED ---- 
    [img_blob]: Returns the final image after filling-in the identified
        blobs. img_blob is in the same format as img_in.

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """
    
    img = img_in.copy()
    blob_params = blob_params_in
    quiet = quiet_in

    area_thresh_min = blob_params[0]
    area_thresh_max = blob_params[1]
    circty_thresh_min = blob_params[2]
    circty_thresh_max = blob_params[3]
    ar_min = blob_params[4]
    ar_max = blob_params[5]
    blob_color = blob_params[6]

    PI = float(3.14159265358979323846264338327950288419716939937510)
    BLACK = (0, 0, 0)
    WHITE = (255, 255, 255)

    area_thresh_min = int(round(area_thresh_min))
    area_thresh_max = int(round(area_thresh_max))

    if area_thresh_min > area_thresh_max:
        area_thresh_min = area_thresh_max
    elif area_thresh_max < area_thresh_min:
        area_thresh_max = area_thresh_min

    if circty_thresh_min > circty_thresh_max:
        circty_thresh_min = circty_thresh_max
    elif circty_thresh_max < circty_thresh_min:
        circty_thresh_max = circty_thresh_min

    if ar_min > ar_max:
        ar_min = ar_max
    elif ar_max < ar_min:
        ar_max = ar_min

    [img_contrs, contr_hierarchy] = cv.findContours(img, cv.RETR_LIST, 
        cv.CHAIN_APPROX_SIMPLE)

    if img_contrs:

        contrs_area = []
        contrs_perim = []
        contrs_circty = []
        contrs_del = []

        for contr in img_contrs:
            cur_area = cv.contourArea(contr)
            contrs_area.append(cur_area)

            cur_perim = cv.arcLength(contr, True)
            contrs_perim.append(cur_perim)

            if cur_perim != 0:
                cur_circty = 4.0*PI*cur_area/(cur_perim*cur_perim)
            else:
                cur_circty = 0.0

            # Could also get the aspect ratio and compare against that
            [x_rect, y_rect, w_rect, h_rect] = cv.boundingRect(contr)
            cur_ar = float(w_rect)/h_rect 

            if (area_thresh_min <= cur_area <= area_thresh_max) and\
                (circty_thresh_min <= cur_circty <= circty_thresh_max) and\
                (ar_min <= cur_ar <= ar_max):
             
                contrs_del.append(contr)        

    # Fill the white blobs with black pixels, effectively deleting them
    if contrs_del: 
        for contr in contrs_del:
            cv.drawContours(img, [contr], 0, blob_color, thickness=cv.FILLED, 
                lineType=cv.LINE_8)

    num_del_blob = len(contrs_del)

    if not quiet:
        print(f"\nSuccessfully filled in {num_del_blob} blobs using the "\
            "following criterion:\n"\
            f"    Min. Area Threshold: {area_thresh_min} pixels\n"\
            f"    Max. Area Threshold: {area_thresh_max} pixels\n"\
            f"    Min. Circularity: {circty_thresh_min}\n"\
            f"    Max. Circularity: {circty_thresh_max}\n"\
            f"    Min. Aspect Ratio: {ar_min}\n"\
            f"    Max. Aspect Ratio: {ar_max}")

    return img


def interact_driver_denoise(img_in):
    """
    Create an interactive session that changes the parameters necessary
    for the OpenCV's advanced non-local means denoising algorithm. The
    user can change the procedure's parameters and get updates in real-
    time. However, this denoising procedure is very computationally
    expensive, so the window may stall for short periods of time with
    large patch sizes. Compared to blur filters, like Gaussian blur,
    this procedure usually does a better job at removing noise while
    maintaining the integrity of the original signal. So, in a sense,
    this procedure is edge preserving, or at least, edge friendly. Upon
    hitting Enter, the interactive window closes, and the final image is
    returned. This function is related to apply_driver_denoise(...).

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    ---- RETURNED ---- 
    [img_denoise]: Returns the final image after closing the interactive
        session. img_denoise is in the same format as img_in

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    [img_denoise, denoise_params] = ifun.interact_denoise(img_in)

    # Using this function just to write to standard output
    apply_driver_denoise(img_in, denoise_params)

    return img_denoise


def apply_driver_denoise(img_in, denoise_params_in, quiet_in=False):
    """
    Implements OpenCV's advanced non-local means denoising algorithm
    like interact_driver_denoise(...), but as a batch process rather 
    than being interactive. This procedure is rather computationally 
    intensive, particularly for large patch sizes. However, Compared to
    blur filters, like Gaussian blur, this procedure usually does a
    better job at removing noise while maintaining the integrity of the
    original signal. So, in a sense, this procedure is edge preserving,
    or at least, edge friendly.

    ---- INPUT ARGUMENTS ---- 
    [img_in]: OpenCV's numpy array for a grayscale image. It is assumed
        that the image is already grayscale and of type uint8. The array
        should thus be 2D, where each value represents the intensity for
        each corresponding pixel. 

    [denoise_params_in]: A list of parameters for input to the non-local
        means denoising algorithm. More details on the parameters of
        denoise_params_in is given below,
    
            [cur_h, cur_tsize, cur_wsize]
                
                cur_h: Parameter regulating filter strength. Big cur_h
                    value perfectly removes noise but also removes image
                    details, smaller cur_h value preserves details but 
                    also preserves some noise.
                
                cur_tsize: Size in pixels of the template patch that is
                    used to compute weights. Should be odd.
                
                cur_wsize: Size in pixels of the window that is used to
                    compute weighted average for given pixel. Should be
                    odd. Affects performance linearly.

    quiet_in: A boolean that determiens if this function should print
        any statements to standard output. If False (default), outputs  
        are written. Conversely, if True, outputs are suppressed. This
        is particularly useful in the event of batch processing

    ---- RETURNED ---- 
    [img_denoise]: Returns the final image after applying the non-local
        means denoising algorithm. img_denoise is in the same format as
        img_in

    ---- SIDE EFFECTS ---- 
    Function input arguments are not altered. Nothing is written to the 
    hard drive. This function is a read-only function. It does pop-up
    a new window that visualizes the provided image. Strings are printed
    to standard output.
    """

    img = img_in.copy()
    denoise_params = denoise_params_in
    quiet = quiet_in

    cur_h = denoise_params[0]
    cur_tsize = denoise_params[1]
    cur_wsize = denoise_params[2]

    if cur_h <= 0:
        img_denoise = img

    else:
        # Must be odd for the denoise function
        if cur_tsize % 2 == 0: 
            cur_tsize += -1

        if cur_wsize % 2 == 0: 
            cur_wsize += -1

        img_denoise = cv.fastNlMeansDenoising(img, h=cur_h, 
            templateWindowSize=cur_tsize, searchWindowSize=cur_wsize)

        if not quiet:
            print("\nSuccessfully applied the non-local means denoising"\
                "filter:\n"\
                f"    Filter Strength: {cur_h}\n"\
                f"    Template Patch Size: {cur_tsize}\n"
                f"    Current Window Size: {cur_wsize}")

    return img_denoise



import numpy as np
from scipy import ndimage as ndim
import vtk
import pyvista as pv
from skimage import measure as meas
from skimage.util import img_as_ubyte
from skimage.util import img_as_float32

from . import im3_processing as im3
from . import volume_image_processing as vol


def make_vtk_uniform_grid(img_arr_in,  scale_spacing=1.0, 
    output_rect_grid=False):
    """
    Create a vtkImageData class from a 3D image sequence using PyVista's
    UniformGrid wrapper. Use PyVista UniformGrid plot() function to 
    visualize the uniform grid model in 3D. To save it to a VTK file,
    use the PyVista save() function. Note, the intensity values will
    be saved in the uniform grid under the name "values".

    ---- INPUT ARGUMENTS ---- 
    [[[img_arr_in]]]: A 3D Numpy array representing the image sequence
        It is important that this is a Numpy array and not a Python list
        of Numpy matrices. The shape of img_arr_in is expected to be as
        (num_images, num_pixel_rows, num_pixel_cols). Note, when
        converting this to spatial coordinates, num_images is the
        Z-component, num_pixel_rows is the Y-component, and
        num_pixel_cols is the X-component. It is expected that the
        images are single-channel, and the data should be of type uint8.
        The intensity values will be assigned to each voxel in the 
        structured grid, so grayscale intensities could be used or a
        binarized image sequence could be used.

    scale_spacing: A scalar float that represents the distance between
        the voxels before they are converted to a surface. If this is
        an X-ray CT image sequence, this represents the voxel size.

    output_rect_grid: A boolean to determine whether a 
        pyvista.RectilinearGrid object should be returned instead of
        a pyvista.UniformGrid object. The pyvista.RectilinearGrid object
        is a wrapper for a vtk.vtkRectilinearGrid object.

    ---- RETURNED ---- 
    uni_grid: A pyvista.UniformGrid object is returned. PyVista is a 
        Python library that acts as a high-level wrapper for VTK - the
        visualization toolkit. PyVista's UniformGrid object is an 
        extension of VTK's vtkImageData class. Note, if output_rect_grid
        is True, then a pyvista.RectilinearGrid object is returned. A
        pyvista.RectilinearGrid object is wrapper for a 
        vtk.vtkRectilinearGrid object. For more information, see
        https://docs.pyvista.org/core/grids.html

    ---- SIDE EFFECTS ---- 
    Function input arguments should not be altered. However, a new view
    is created for img_arr_in, not a deep copy. So, this cannot be 
    guaranteed. Nothing is written to the hard drive.
    """

    img_arr = np.transpose(img_arr_in) # New view, but not deep copy

    # Create the spatial reference
    uni_grid = pv.UniformGrid()

    # Set the grid dimensions: shape + 1 because we want to inject the values on
    # the CELL data
    uni_grid.dimensions = np.array(img_arr.shape) + 1

    # Translate the origin of the mesh
    #grid.origin = (100, 33, 55.6)

    # These are the cell sizes along each axis. Uniform/unity spacing
    uni_grid.spacing = (scale_spacing, scale_spacing, scale_spacing)  

    # Add the data values to the cell data
    # Flatten the array (in Fortran style)
    uni_grid.cell_data["values"] = img_arr.flatten(order="F") 

    if output_rect_grid:
        uni_grid = uni_grid.cast_to_rectilinear_grid()

    return uni_grid


def convert_voxels_to_surface(img_arr_in, iso_level=125, scale_spacing=1.0,
    is_binary=True, g_sigdev=0.8, pad_boundary=True):
    """
    Applies a marching cubes algorithm to an image sequence (i.e., a
    voxel model) in order to calculate an isosurface. The algorithm used
    is the SciKit-Image algorithm, skimage.measure.marching_cubes(). It
    is based on the Lewiner et al. approach:

    Thomas Lewiner, Helio Lopes, Antonio Wilson Vieira and Geovan Tavares.
      Efficient implementation of Marching Cubes’ cases with topological 
      guarantees. Journal of Graphics Tools 8(2) pp. 1-15 (December 2003).
      DOI:10.1080/10867651.2003.10487582

    No additional smoothing or modification to the resultant mesh is 
    performed. If this is desired, then consider making a VTK surface
    mesh and using Laplacian smoothing, which is all available with 
    little effort in make_vtk_surf_mesh() defined below. To measure
    the surface area, use skimage.measure.mesh_surface_area(). Or, 
    create a VTK surface mesh using make_vtk_surf_mesh() and then
    retrieve the pyvista.PolyData.area property.

    ---- INPUT ARGUMENTS ---- 
    [[[img_arr_in]]]: A 3D Numpy array representing the image sequence
        It is important that this is a Numpy array and not a Python list
        of Numpy matrices. The shape of img_arr_in is expected to be as
        (num_images, num_pixel_rows, num_pixel_cols). Note, when
        converting this to spatial coordinates, num_images is the
        Z-component, num_pixel_rows is the Y-component, and
        num_pixel_cols is the X-component. It is expected that the
        images are single-channel, and the data should be of type uint8.
        Specifically, intensities should range from 0 to 255. If it is
        segmented (i.e., binarized), then only values of 0 and 255 
        should be present (NOT 0 and 1).

    iso_level: A scalar number used to define the value for the iso-
        surface. For a binarized image sequence, a value in between 
        0 and 255, like 125, is recommended. For grayscale image
        sequence, any value can be used, and the corresponding 
        isosurface will be approximated.

    scale_spacing: A scalar float that represents the distance between
        the voxels before they are converted to a surface. If this is
        an X-ray CT image sequence, this represents the voxel size.

    is_binary: Set to True if the image sequence is binarized. That is,
        if it contains only values of 0 and 255. Using the binaraized
        image sequence directly into the marching cubes algorithm will
        result in a very poor isosurface. So, if is_binary is True, 
        then a small amount of Gaussian blurring is performed which makes 
        for a much smoother gradient between black and white. This
        vastly improves the accuracy of the marching cubes algorithm.

    g_sigdev: A single float that corresponds to the standard deviation
        used as input for the 3D Gaussian blur that must be applied for
        a binary image sequence. If is_binary is False, this value is
        ignored. The same standard deviation will be used for all three
        axes. Typical values range between 0.5 and 1.5. This blur helps
        to stabilize the marching cubes algorithm, but it also has the
        side effect of smoothing edges by a small amount.

    pad_boundary: Set to True to extend the boundaries of the image
        sequence in all three dimensions. More specifically, in all
        six directions of the rectangular volume defined by the image
        sequence. The six boundaries will be extended by some amount. 
        The motivation to do this is to ensure
        there are no white pixels along the boundaries of the image
        sequence. Otherwise, the resultant surface will not be a 
        closed at the boundaries. If there are no white pixels along the
        boundaries to begin with, then there is no need for padding.
        
    ---- RETURNED ----
    [[verts]]: A 2D array that contains the coordinates of the mesh 
        vertices. The array will have 3 columns, such that each row
        will provide the [X, Y, Z] coordinates of that vertex. There
        will be V unique mesh vertices. So, the size is (V, 3).

    [[faces]]: A 2D array that defines the connectivity of the 
        triangular faces. There will be F faces, and for each face,
        three nodes are defined based on the indices of verts. So, the
        size is (F, 3).

    [[normals]]: A 2D array that defines the normal vectors at each
        vertex. So, the size is (V, 3). Each normal vector contains
        a [X, Y, Z] components.

    [values]: A 1D array containing the approximate pixel values
        at each vertex. So, the size is (1, V)

    ---- SIDE EFFECTS ---- 
    Function input arguments should not be altered. However, a new view
    is created for img_arr_in, not a deep copy. So, this cannot be 
    guaranteed. Nothing is written to the hard drive.
    """

    img_arr = np.transpose(img_arr_in) # New view, but not deep copy

    if is_binary:
        # Blurring something white at the edge does not create a 
        # smooth gradient to black. So, pad the edges with black
        # pixels by making the image sequence larger.
        img_arr = vol.pad_image_boundary(img_arr, n_pad_in=9, quiet_in=True)

        # Apply a little blur to get better isosurfaces on a binary
        # image sequence. Need smoother gradients
        g_sigma = (g_sigdev, g_sigdev, g_sigdev) 
        img_arr = img_as_float32(img_arr)
        img_arr = ndim.gaussian_filter(img_arr, g_sigma)
        img_arr = img_as_ubyte(img_arr)

    if pad_boundary:
        # If an isosurface comes to the edge, the resultant mesh is left
        # with an open hole. By padding the image array with black pixels
        # (by making it a little larger), the isosurface will be closed.
        img_arr = vol.pad_image_boundary(img_arr, n_pad_in=1, quiet_in=True)

    # Apply marching cubes algorithm
    # verts is a (V,3) array. 
    # faces is a (F,3) array.
    # normals is a (V,3) array.
    # values is a (V,) array
    verts, faces, normals, values = meas.marching_cubes(img_arr, 
        level=iso_level, spacing=(scale_spacing, scale_spacing, scale_spacing), 
        gradient_direction='descent', step_size=1, 
        allow_degenerate=False, method='lewiner')

    # Offset the vertices by the amount of padding
    if is_binary:
        verts[:,0] = verts[:,0] - 9*scale_spacing
        verts[:,1] = verts[:,1] - 9*scale_spacing
        verts[:,2] = verts[:,2] - 9*scale_spacing

    if pad_boundary:
        verts[:,0] = verts[:,0] - 1*scale_spacing
        verts[:,1] = verts[:,1] - 1*scale_spacing
        verts[:,2] = verts[:,2] - 1*scale_spacing

    return [verts, faces, normals, values]


def make_vtk_surf_mesh(verts_in, faces_in, values_in, smth_iter=5):
    """
    Creates a PyVista PolyData object, which is an extension of the
    vtk.vtkPolyData object. Based on a set of nodal coordinates and
    the connectivity of these nodes to form triangular faces, and 
    surface mesh object is created. Use PyVista PolyData plot() 
    function to  visualize the mesh model in 3D. To save it to a VTK 
    file, use the PyVista save() function. Laplacian smoothing can also
    be applied the surface mesh before it is returned. The inputs
    for this function are based on the outputs of 
    convert_voxels_to_surface() defined above.

    ---- INPUT ARGUMENTS ----
    [[verts_in]]: A 2D array that contains the coordinates of the mesh 
        vertices. The array should have 3 columns, such that each row
        will provide the [X, Y, Z] coordinates of that vertex. Assuming
        there are V unique mesh vertices, then the size is (V, 3).

    [[faces_in]]: A 2D array that defines the connectivity of the 
        triangular faces. Assuming there are F faces, then each face 
        will require three nodes to define it. These indices of the
        desired nodes from verts_in are used to define this 
        connectivity. The size of faces_in should be (F, 3).

    [[values_in]]: A 1D array corresponding to the values that should
        be assigned to each vertex. These will be saved as a point_array
        in the pyvista.PolyData object, called "values".

    smth_iter: A scalar integer that defines the number of iterations
        of Laplacian smoothing to be applied. A value of 0 will prevent
        any smoothing from occurring. 

    ---- RETURNED ----
    surf: A PyVista PolyData object that represents the triangular mesh.
        For more information, go to:

        https://docs.pyvista.org/core/points.html?highlight=polydata#
        pyvista.PolyData

    ---- SIDE EFFECTS ---- 
    Function input arguments should not be altered. However, a new view
    is created for the vertices and faces. So, this cannot be guaranteed. 
    Nothing is written to the hard drive.
    """

    verts = verts_in
    faces = faces_in
    values = values_in
    smth_iter = np.round(smth_iter)

    num_faces = faces.shape[0]
    vec_of_3s = (np.ones((num_faces,1))*3)
    faces2 = np.hstack( np.column_stack((vec_of_3s,faces)) ).astype(np.uint32)
    surf = pv.PolyData(verts, faces2)

    if smth_iter > 0:
        surf = surf.smooth(n_iter=smth_iter, inplace=False)

    surf.point_data["values"] = values
    surf.compute_normals(inplace=True)

    return surf


def make_vtk_unstruct_grid_slow(img_arr_in):
    """
    This function converts an image sequence (i.e., a 3D array that
    contains 2D images) into an unstructured VTK voxel model. Although
    this function creates the voxel model in an efficient manner
    concerning RAM, it is super slow thanks to some big Python 
    for-loops. Best suited for models that are less than 128 by 128 by
    128 pixels in total size. For a faster implementation based on a
    custom C-extension, see make_vtk_unstruct_grid() below. Note, this
    function will only convert pixels with intensity values greater than
    or equal to one into voxels -- black pixels are skipped.

     ---- INPUT ARGUMENTS ---- 
    [[[img_arr_in]]]: A 3D Numpy array representing the image sequence
        It is important that this is a Numpy array and not a Python list
        of Numpy matrices. The shape of img_arr_in is expected to be as
        (num_images, num_pixel_rows, num_pixel_cols). Note, when
        converting this to spatial coordinates, num_images is the
        Z-component, num_pixel_rows is the Y-component, and
        num_pixel_cols is the X-component. It is expected that the
        images are single-channel, and the data should be of type uint8.
        Specifically, intensities should range from 0 to 255. If it is
        segmented (i.e., binarized), then only values of 0 and 255 
        should be present (NOT 0 and 1).

    ---- RETURNED ----
    unstruct_grid: A pyvista.UnstructuredGrid object is returned. 
        PyVista is a Python library that acts as a high-level wrapper
        for VTK - the visualization toolkit. PyVista's UnstructuredGrid
        object is an extension of VTK's vtkUnstructuredGrid class. For
        more information, see 
        https://docs.pyvista.org/core/point-grids.html

    ---- SIDE EFFECTS ----
    Function input arguments should not be altered. However, a new view
    is created for img_arr_in, not a deep copy. So, this cannot be 
    guaranteed. Nothing is written to the hard drive.
    """

    img_arr = img_arr_in # Do not transpose here!
    img_arr_shape = img_arr.shape
    n_imgs = img_arr_shape[0]
    n_rows = img_arr_shape[1]
    n_cols = img_arr_shape[2]

    # The number of voxels that will be created. Initialize the Numpy
    # connectivity array that will be used to create the VTK model
    num_cells = np.count_nonzero(img_arr >= 1)
    cells = np.zeros((num_cells,9), dtype=np.uint32)
    vals = np.ones(num_cells, dtype=np.uint8)*255
    vert_ids = np.ones((n_imgs+1, n_rows+1, n_cols+1), dtype=np.uint32)\
        *4294967295 # Largest possible integer in a uint32 data type

    node_arr = []
    nd_k = 0 # Node IDs start at zero
    cell_k = 0 # Cell IDs start at zero
    for ii in range(0, n_imgs):
        for rr in range(0, n_rows):
            for cc in range(0, n_cols):

                cur_int = img_arr[ii,rr,cc]
                if cur_int < 1:
                    continue # Skip black voxels

                xc = cc # Converting to XYZ coordinates here
                yc = rr
                zc = ii

                # Construct the node array spatial coordinates 
                nd1 = np.array([xc-0.5, yc-0.5, zc-0.5]).astype(np.float32)
                nd2 = np.array([xc+0.5, yc-0.5, zc-0.5]).astype(np.float32)
                nd3 = np.array([xc+0.5, yc+0.5, zc-0.5]).astype(np.float32)
                nd4 = np.array([xc-0.5, yc+0.5, zc-0.5]).astype(np.float32)
                nd5 = np.array([xc-0.5, yc-0.5, zc+0.5]).astype(np.float32)
                nd6 = np.array([xc+0.5, yc-0.5, zc+0.5]).astype(np.float32)
                nd7 = np.array([xc+0.5, yc+0.5, zc+0.5]).astype(np.float32)
                nd8 = np.array([xc-0.5, yc+0.5, zc+0.5]).astype(np.float32)

                if vert_ids[ii, rr, cc] == 4294967295: # Node 1 not being used
                    vert_ids[ii, rr, cc] = nd_k
                    nd1_id = nd_k
                    node_arr.append(nd1)
                    nd_k += 1
                else:
                    nd1_id = vert_ids[ii, rr, cc]

                if vert_ids[ii, rr, cc+1] == 4294967295: # Node 2 not being used
                    vert_ids[ii, rr, cc+1] = nd_k
                    nd2_id = nd_k
                    node_arr.append(nd2)
                    nd_k += 1
                else:
                    nd2_id = vert_ids[ii, rr, cc+1]

                if vert_ids[ii, rr+1, cc+1] == 4294967295: # Node 3 not being used
                    vert_ids[ii, rr+1, cc+1] = nd_k
                    nd3_id = nd_k
                    node_arr.append(nd3)
                    nd_k += 1
                else:
                    nd3_id = vert_ids[ii, rr+1, cc+1]

                if vert_ids[ii, rr+1, cc] == 4294967295: # Node 4 not being used
                    vert_ids[ii, rr+1, cc] = nd_k
                    nd4_id = nd_k
                    node_arr.append(nd4)
                    nd_k += 1
                else:
                    nd4_id = vert_ids[ii, rr+1, cc]

                if vert_ids[ii+1, rr, cc] == 4294967295: # Node 5 not being used
                    vert_ids[ii+1, rr, cc] = nd_k
                    nd5_id = nd_k
                    node_arr.append(nd5)
                    nd_k += 1
                else:
                    nd5_id = vert_ids[ii+1, rr, cc]

                if vert_ids[ii+1, rr, cc+1] == 4294967295: # Node 6 not being used
                    vert_ids[ii+1, rr, cc+1] = nd_k
                    nd6_id = nd_k
                    node_arr.append(nd6)
                    nd_k += 1
                else:
                    nd6_id = vert_ids[ii+1, rr, cc+1]

                if vert_ids[ii+1, rr+1, cc+1] == 4294967295: # Node 7 not being used
                    vert_ids[ii+1, rr+1, cc+1] = nd_k
                    nd7_id = nd_k
                    node_arr.append(nd7)
                    nd_k += 1
                else:
                    nd7_id = vert_ids[ii+1, rr+1, cc+1]

                if vert_ids[ii+1, rr+1, cc] == 4294967295: # Node 8 not being used
                    vert_ids[ii+1, rr+1, cc] = nd_k
                    nd8_id = nd_k
                    node_arr.append(nd8)
                    nd_k += 1
                else:
                    nd8_id = vert_ids[ii+1, rr+1, cc]

                cells[cell_k,:] = np.array([8, nd1_id, nd2_id, nd3_id, nd4_id,\
                    nd5_id, nd6_id, nd7_id, nd8_id])

                vals[cell_k] = cur_int

                cell_k += 1

    del vert_ids # No longer need this 

    # Convert this back to a 2D numpy array. No need to keep the values. 
    # The indices themselves for each row are the values (i.e., node IDs).
    #coords = np.array(list(coords.keys()))
    node_arr = np.array(node_arr)

    # Different ways to initialize the grid if using VTK version 9.0 or newer
    vtk_ver_junk_str = vtk.vtkVersion.GetVTKSourceVersion()
    vtk_ver_str = ''.join(c for c in vtk_ver_junk_str if c.isdigit())
    vtk_ver_num = int(vtk_ver_str[0])

    if vtk_ver_num >= 9: 
        # I don't have Version 9 of VTK, so this is untested
        # Using PyVista example for this code sample:
        #
        # https://docs.pyvista.org/examples/00-load/create-unstructured-
        # surface.html#sphx-glr-examples-00-load-create-unstructured-surface-py

        unstruct_grid = pv.UnstructuredGrid({vtk.VTK_HEXAHEDRON: cells[:, 1:]}, 
            node_arr)

    else:
        # Flattens this into a 1D-array (using C-like row-major order)
        cells = cells.ravel()

        # Offset/index index to go to each cell for the flattened array
        off_end = num_cells*9
        offset = np.arange(0, off_end, 9, dtype=np.uint32)

        # Each cell is a VTK_HEXAHEDRON
        cell_types = np.empty(num_cells, dtype=np.uint8)
        cell_types[:] = vtk.VTK_HEXAHEDRON

        unstruct_grid = pv.UnstructuredGrid(offset, cells, cell_types, node_arr)

    # Assign the intensity values to the cells
    unstruct_grid.cell_data["values"] = vals

    return unstruct_grid


def make_vtk_unstruct_grid(img_arr_in, scale_spacing=1.0, 
    all_voxels=False, quiet_in=False):
    """
    This function converts an image sequence into an unstructured VTK
    voxel model. This function depends on a C-extension binary that
    was created using Cython. This function is the super fast version
    of make_vtk_unstruct_grid_slow(), and in general, is preferred.
    However, the binary file, called im3_processing.cp37-win_amd64.pyd
    was compiled for Python 3.7 for Windows x64. If you are using a 
    different operating system or version of Python, then the .pyx
    file will need to be re-compiled for a new target. Contact Newell
    Moser for assistance if this is the case. Otherwise, you will need
    to re-compile it or use make_vtk_unstruct_grid_slow(). By default,
    this function will only convert pixels with intensity values greater
    than or equal to one -- black pixels are skipped. However, if RAM
    is not an issue, then set all_voxels equal to True to convert all
    of the pixels. Use PyVista UniformGrid plot() function to 
    visualize the uniform grid model in 3D. To save it to a VTK file,
    use the PyVista save() function. Note, the intensity values will
    be saved in the uniform grid under the name "values".

     ---- INPUT ARGUMENTS ---- 
    [[[img_arr_in]]]: A 3D Numpy array representing the image sequence
        It is important that this is a Numpy array and not a Python list
        of Numpy matrices. The shape of img_arr_in is expected to be as
        (num_images, num_pixel_rows, num_pixel_cols). Note, when
        converting this to spatial coordinates, num_images is the
        Z-component, num_pixel_rows is the Y-component, and
        num_pixel_cols is the X-component. It is expected that the
        images are single-channel, and the data should be of type uint8.
        Specifically, intensities should range from 0 to 255. If it is
        segmented (i.e., binarized), then only values of 0 and 255 
        should be present (NOT 0 and 1).

    scale_spacing: A scalar float that represents the distance between
        the voxels before they are converted to a surface. If this is
        an X-ray CT image sequence, this represents the voxel size.

    all_voxels: By default, this is set to False, which means only non-
        black pixels will be converted to voxels. Specifically, pixels
        with intensites greater than or equal to one. Consequently, 
        this function will return a pyvista.UnstructuredGrid object. Set
        this to True in order to return all of the pixels as voxels, 
        in which case, a pyvista.StructuredGrid object is returned.

    quiet_in: An optional boolean that is by default set to False. Set 
        to True to prevent outputting any messages. 

    ---- RETURNED ----
    unstruct_grid: If all_voxels is False, a pyvista.UnstructuredGrid 
        object is returned. PyVista is a Python library that acts as a
        high-level wrapper for VTK - the visualization toolkit. 
        PyVista's UnstructuredGrid object is an extension of VTK's 
        vtkUnstructuredGrid class. If If all_voxels is True, then a 
        pyvista.StructuredGrid object is returned. PyVista's 
        StructuredGrid object is an extension of VTK's 
        vtk.vtkStructuredGrid object. For more information, see 
        https://docs.pyvista.org/core/point-grids.html

    ---- SIDE EFFECTS ----
    Function input arguments should not be altered. However, a new view
    is created for img_arr_in, not a deep copy. So, this cannot be 
    guaranteed. Nothing is written to the hard drive.
    """

    # Do NOT transpose here. My C-extension expects the image array
    # in image coordinates, and will return it correctly in XYZ.
    img_arr = img_arr_in 

    n_imgs = img_arr_in.shape[0]
    n_rows = img_arr_in.shape[1]
    n_cols = img_arr_in.shape[2]

    quiet = quiet_in

    if all_voxels: # Without downsampling, this WILL kill your RAM
        if not quiet:
            print("\nAttempting to construct the structured VTK file...")

        unstruct_grid = make_vtk_uniform_grid(img_arr)
        unstruct_grid = unstruct_grid.cast_to_structured_grid()
        
        if not quiet:
            print("  Success!")

    else:
        # Call my Cython C-extension. Set the second argument, "debug_flag",
        # to 1 (instead of 0) if you want the expected memory allocations to 
        # also be written to the terminal standard output.
        if not quiet:
            print("\nAllocating memory for the VTK voxel data structure...")
        
        nd_coords, cell_conn, cell_vals = im3.get_voxel_info(img_arr, 0)
        num_cells = cell_conn.shape[0]

        nd_coords = nd_coords*scale_spacing

        # Different ways to initialize the grid if using VTK version 9.0 or newer
        vtk_ver_junk_str = vtk.vtkVersion.GetVTKSourceVersion()
        vtk_ver_str = ''.join(c for c in vtk_ver_junk_str if c.isdigit())
        vtk_ver_num = int(vtk_ver_str[0])

        if not quiet:
            print("\nAttempting to construct the unstructured VTK file...")
        
        if vtk_ver_num >= 9: 
            # I don't have Version 9 of VTK, so this is untested
            # Using a PyVista example for this code sample:
            #
            # https://docs.pyvista.org/examples/00-load/create-unstructured-
            # surface.html#sphx-glr-examples-00-load-create-unstructured-surface-py

            unstruct_grid = pv.UnstructuredGrid(
                {vtk.VTK_HEXAHEDRON: cell_conn[:, 1:]}, nd_coords)

        else:
            # Flattens this into a 1D-array (using C-like row-major order)
            cell_conn = cell_conn.ravel()

            # Offset/index index to go to each cell for the flattened array
            off_end = num_cells*9
            offset = np.arange(0, off_end, 9, dtype=np.uint32)

            # Each cell is a VTK_HEXAHEDRON
            cell_types = np.empty(num_cells, dtype=np.uint8)
            cell_types[:] = vtk.VTK_HEXAHEDRON

            unstruct_grid = pv.UnstructuredGrid(offset, cell_conn, 
                cell_types, nd_coords)

        # Assign the intensity values to the cell_conn
        unstruct_grid.cell_data["values"] = cell_vals

    if not quiet:
        print("  Success!")

    return unstruct_grid
